import unittest

import jax
jax.config.update("jax_enable_x64", True)
import jax.random as random
import jax.numpy as jnp

import numpy as np

import sys
sys.path.append(sys.path[0] + '/../')

import jVMC
import jVMC.nets as nets
from jVMC.vqs import NQS
import jVMC.sampler as sampler
import jVMC.mpi_wrapper as mpi


def state_to_int(s):

    def for_fun(i, xs):
        return (xs[0] + xs[1][i] * (2**i), xs[1])

    x, _ = jax.lax.fori_loop(0, s.shape[-1], for_fun, (0, s))

    return x


class TestMC(unittest.TestCase):

    def test_MCMC_sampling(self):
        L = 4

        weights = jnp.array(
            [0.23898957, 0.12614753, 0.19479055, 0.17325271, 0.14619853, 0.21392751,
             0.19648707, 0.17103704, -0.15457255, 0.10954413, 0.13228065, -0.14935214,
             -0.09963073, 0.17610707, 0.13386381, -0.14836467]
        )

        # Set up variational wave function
        rbm = nets.CpxRBM(numHidden=2, bias=False)
        orbit = jVMC.util.symmetries.get_orbit_1D(L, "translation", "reflection", "spinflip")
        net = nets.sym_wrapper.SymNet(net=rbm, orbit=orbit)
        psi = NQS(net)

        # Set up exact sampler
        exactSampler = sampler.ExactSampler(psi, L)

        # Set up MCMC sampler
        mcSampler = sampler.MCSampler(psi, (L,), random.PRNGKey(0), updateProposer=jVMC.sampler.propose_spin_flip, numChains=777)

        p0 = psi.get_parameters()
        psi.set_parameters(weights)

        # Compute exact probabilities
        _, _, pex = exactSampler.sample()
        pex = mpi.gather(pex)

        # Get samples from MCMC sampler
        numSamples = 500000
        smc, _, p = mcSampler.sample(numSamples=numSamples)

        smc = smc.reshape((smc.shape[0] * smc.shape[1], -1))

        self.assertTrue(mpi.global_sum(jnp.array([smc.shape[0],])[None,None,...]) >= numSamples)

        # Compute histogram of sampled configurations
        smcInt = jax.vmap(state_to_int)(smc)
        pmc, _ = np.histogram(smcInt, bins=np.arange(0, 17), weights=p[0])
        pmc = mpi.global_sum(jnp.array(pmc)[None,None,...])
        pmc = pmc / jnp.sum(pmc)

        # Compare histogram to exact probabilities
        self.assertTrue(jnp.max(jnp.abs(pmc - pex.reshape((-1,))[:16])) < 2e-3)

        
        s, psi_s, _ = mcSampler.sample(parameters=p0, numSamples=100)
        psi.set_parameters(p0)
        psi_s1 = psi(s)
        
        self.assertTrue(jnp.max(jnp.abs((psi_s - psi_s1) / psi_s)) < 1e-14)


    def test_MCMC_sampling_with_mu(self):
        mu = 1

        L = 4

        weights = jnp.array(
            [0.23898957, 0.12614753, 0.19479055, 0.17325271, 0.14619853, 0.21392751,
             0.19648707, 0.17103704, -0.15457255, 0.10954413, 0.13228065, -0.14935214,
             -0.09963073, 0.17610707, 0.13386381, -0.14836467]
        )

        # Set up variational wave function
        rbm = nets.CpxRBM(numHidden=2, bias=False)
        orbit = jVMC.util.symmetries.get_orbit_1D(L)
        net = nets.sym_wrapper.SymNet(net=rbm, orbit=orbit)
        psi = NQS(net)

        # Set up exact sampler
        exactSampler = sampler.ExactSampler(psi, L)

        # Set up MCMC sampler
        mcSampler = sampler.MCSampler(psi, (L,), random.PRNGKey(0), updateProposer=jVMC.sampler.propose_spin_flip, numChains=777, mu=mu)

        psi.set_parameters(weights)

        # Compute exact probabilities
        _, _, pex = exactSampler.sample()
        pex = mpi.gather(pex)

        # Get samples from MCMC sampler
        numSamples = 500000
        smc, logPsi, weighting_probs = mcSampler.sample(numSamples=numSamples)

        smc = smc.reshape((smc.shape[0] * smc.shape[1], -1))

        self.assertTrue(mpi.global_sum(jnp.array([smc.shape[0],])[None,None,...]) >= numSamples)

        # Compute histogram of sampled configurations
        smcInt = jax.vmap(state_to_int)(smc)

        psi_log_basis = psi(exactSampler.basis)

        pmc, _ = np.histogram(smcInt, bins=np.arange(0, 17), weights=weighting_probs[0, :])
        pmc = mpi.global_sum(jnp.array(pmc)[None,None,...])
        pmc = pmc / jnp.sum(pmc)

        # Compare histogram to exact probabilities
        self.assertTrue(jnp.max(jnp.abs(pmc - pex.reshape((-1,))[:16])) < 2e-3)

    def test_MCMC_sampling_with_logProbFactor(self):
        mu = 1
        logProbFactor = 1
        L = 4

        weights = jnp.array(
            [0.23898957, 0.12614753, 0.19479055, 0.17325271, 0.14619853, 0.21392751,
             0.19648707, 0.17103704, -0.15457255, 0.10954413, 0.13228065, -0.14935214,
             -0.09963073, 0.17610707, 0.13386381, -0.14836467]
        )

        # Set up variational wave function
        rbm = nets.CpxRBM(numHidden=2, bias=False)
        orbit = jVMC.util.symmetries.get_orbit_1D(L)
        net = nets.sym_wrapper.SymNet(net=rbm, orbit=orbit)
        psi = NQS(net)

        # Set up exact sampler
        exactSampler = sampler.ExactSampler(psi, L, logProbFactor=logProbFactor)

        # Set up MCMC sampler
        mcSampler = sampler.MCSampler(psi, (L,), random.PRNGKey(0), updateProposer=jVMC.sampler.propose_spin_flip,
                                      numChains=777, mu=mu, logProbFactor=logProbFactor)

        psi.set_parameters(weights)

        # Compute exact probabilities
        _, _, pex = exactSampler.sample()
        pex = mpi.gather(pex)

        # Get samples from MCMC sampler
        numSamples = 500000
        smc, logPsi, weighting_probs = mcSampler.sample(numSamples=numSamples)

        smc = smc.reshape((smc.shape[0] * smc.shape[1], -1))

        self.assertTrue(mpi.global_sum(jnp.array([smc.shape[0],])[None,None,...]) >= numSamples)

        # Compute histogram of sampled configurations
        smcInt = jax.vmap(state_to_int)(smc)

        psi_log_basis = psi(exactSampler.basis)

        pmc, _ = np.histogram(smcInt, bins=np.arange(0, 17), weights=weighting_probs[0, :])
        pmc = mpi.global_sum(jnp.array(pmc)[None,None,...])
        pmc = pmc / jnp.sum(pmc)

        # Compare histogram to exact probabilities
        self.assertTrue(jnp.max(jnp.abs(pmc - pex.reshape((-1,))[:16])) < 2e-3)

    def test_MCMC_sampling_with_two_nets(self):
        L = 4

        weights = jnp.array(
            [0.23898957, 0.12614753, 0.19479055, 0.17325271, 0.14619853, 0.21392751,
             0.19648707, 0.17103704, -0.15457255, 0.10954413, 0.13228065, -0.14935214,
             -0.09963073, 0.17610707, 0.13386381, -0.14836467]
        )

        # Set up variational wave function
        rbm1 = nets.RBM(numHidden=2, bias=False)
        rbm2 = nets.RBM(numHidden=2, bias=False)
        model = jVMC.nets.two_nets_wrapper.TwoNets((rbm1, rbm2))
        orbit = jVMC.util.symmetries.get_orbit_1D(L)
        net = nets.sym_wrapper.SymNet(net=model, orbit=orbit)
        psi = NQS(net)

        # Set up exact sampler
        exactSampler = sampler.ExactSampler(psi, L)

        # Set up MCMC sampler
        mcSampler = sampler.MCSampler(psi, (L,), random.PRNGKey(0), updateProposer=jVMC.sampler.propose_spin_flip, numChains=777)

        psi.set_parameters(jnp.concatenate([weights, weights]))

        # Compute exact probabilities
        _, _, pex = exactSampler.sample()
        pex = mpi.gather(pex)

        # Get samples from MCMC sampler
        numSamples = 500000
        smc, _, p = mcSampler.sample(numSamples=numSamples)

        smc = smc.reshape((smc.shape[0] * smc.shape[1], -1))

        self.assertTrue(mpi.global_sum(jnp.array([smc.shape[0],])[None,None,...]) >= numSamples)

        # Compute histogram of sampled configurations
        smcInt = jax.vmap(state_to_int)(smc)
        pmc, _ = np.histogram(smcInt, bins=np.arange(0, 17), weights=p[0])
        pmc = mpi.global_sum(jnp.array(pmc)[None,None,...])
        pmc = pmc / jnp.sum(pmc)

        # Compare histogram to exact probabilities
        self.assertTrue(jnp.max(jnp.abs(pmc - pex.reshape((-1,))[:16])) < 2e-3)

    def test_MCMC_sampling_with_integer_key(self):
        L = 4

        weights = jnp.array(
            [0.23898957, 0.12614753, 0.19479055, 0.17325271, 0.14619853, 0.21392751,
             0.19648707, 0.17103704, -0.15457255, 0.10954413, 0.13228065, -0.14935214,
             -0.09963073, 0.17610707, 0.13386381, -0.14836467]
        )

        # Set up variational wave function
        rbm = nets.CpxRBM(numHidden=2, bias=False)
        orbit = jVMC.util.symmetries.get_orbit_1D(L)
        net = nets.sym_wrapper.SymNet(net=rbm, orbit=orbit)
        psi = NQS(net)

        # Set up exact sampler
        exactSampler = sampler.ExactSampler(psi, L)

        # Set up MCMC sampler
        mcSampler = sampler.MCSampler(psi, (L,), 123, updateProposer=jVMC.sampler.propose_spin_flip, numChains=777)

        psi.set_parameters(weights)

        # Compute exact probabilities
        _, _, pex = exactSampler.sample()
        pex = mpi.gather(pex)

        # Get samples from MCMC sampler
        numSamples = 500000
        smc, _, p = mcSampler.sample(numSamples=numSamples)

        smc = smc.reshape((smc.shape[0] * smc.shape[1], -1))

        self.assertTrue(mpi.global_sum(jnp.array([smc.shape[0],])[None,None,...]) >= numSamples)

        # Compute histogram of sampled configurations
        smcInt = jax.vmap(state_to_int)(smc)
        pmc, _ = np.histogram(smcInt, bins=np.arange(0, 17), weights=p[0])
        pmc = mpi.global_sum(jnp.array(pmc)[None,None,...])
        pmc = pmc / jnp.sum(pmc)

        # Compare histogram to exact probabilities
        self.assertTrue(jnp.max(jnp.abs(pmc - pex.reshape((-1,))[:16])) < 2e-3)

    def test_autoregressive_sampling(self):

        L = 4

        # Set up variational wave function
        rnn = nets.RNN1DGeneral(L=4, hiddenSize=5, depth=2)
        rbm = nets.RBM(numHidden=2, bias=False)

        model = jVMC.nets.two_nets_wrapper.TwoNets((rnn, rbm))
        orbit = jVMC.util.symmetries.get_orbit_1D(L)
        net = nets.sym_wrapper.SymNet(net=model, orbit=orbit, avgFun=jVMC.nets.sym_wrapper.avgFun_Coefficients_Sep)
        psi = NQS(net)

        # Set up exact sampler
        exactSampler = sampler.ExactSampler(psi, L)

        # Set up MCMC sampler
        mcSampler = sampler.MCSampler(psi, (L,), random.PRNGKey(0), updateProposer=jVMC.sampler.propose_spin_flip, numChains=777)

        ps = psi.get_parameters()
        psi.update_parameters(ps)

        # Compute exact probabilities
        _, _, pex = exactSampler.sample()
        pex = mpi.gather(pex)

        numSamples = 1000000
        smc, logPsi, p = mcSampler.sample(numSamples=numSamples)

        self.assertTrue(jnp.max(jnp.abs(jnp.real(psi(smc) - logPsi))) < 1e-12)

        smc = smc.reshape((smc.shape[0] * smc.shape[1], -1))

        self.assertTrue(mpi.global_sum(jnp.array([smc.shape[0],])[None,None,...]) >= numSamples)

        # Compute histogram of sampled configurations
        smcInt = jax.vmap(state_to_int)(smc)
        pmc, _ = np.histogram(smcInt, bins=np.arange(0, 17), weights=p[0])
        pmc = mpi.global_sum(jnp.array(pmc)[None,None,...])
        pmc = pmc / jnp.sum(pmc)

        self.assertTrue(jnp.max(jnp.abs(pmc - pex.reshape((-1,))[:16])) < 1.1e-3)
        
        
        psi1 = NQS(net, seed=98475)
        # Set up another MCMC sampler
        mcSampler = sampler.MCSampler(psi1, (L,), random.PRNGKey(0), updateProposer=jVMC.sampler.propose_spin_flip, numChains=777)
        s, psi_s, _ = mcSampler.sample(parameters=psi.get_parameters(), numSamples=100)
        psi_s1 = psi(s)
        
        self.assertTrue(jnp.max(jnp.abs((psi_s - psi_s1) / psi_s)) < 1e-14)

    def test_autoregressive_sampling_with_symmetries(self):

        L = 4

        # Set up variational wave function
        rnn = nets.RNN1DGeneral(L=L, hiddenSize=5, realValuedOutput=True)
        rbm = nets.RBM(numHidden=2, bias=False)
        orbit = jVMC.util.symmetries.get_orbit_1D(L, "translation")
        psi = NQS((rnn, rbm), orbit=orbit, avgFun=jVMC.nets.sym_wrapper.avgFun_Coefficients_Sep)

        # Set up exact sampler
        exactSampler = sampler.ExactSampler(psi, L)

        # Set up MCMC sampler
        mcSampler = sampler.MCSampler(psi, (L,), random.PRNGKey(0), numChains=777, updateProposer=jVMC.sampler.propose_spin_flip)

        # Compute exact probabilities
        _, logPsi, pex = exactSampler.sample()
        pex = mpi.gather(pex)

        numSamples = 1000000
        smc, logPsi, p = mcSampler.sample(numSamples=numSamples)

        self.assertTrue(jnp.max(jnp.abs(jnp.real(psi(smc) - logPsi))) < 1e-12)

        smc = smc.reshape((smc.shape[0] * smc.shape[1], -1))

        self.assertTrue(mpi.global_sum(jnp.array([smc.shape[0],])[None,None,...]) >= numSamples)

        # Compute histogram of sampled configurations
        smcInt = jax.vmap(state_to_int)(smc)
        pmc, _ = np.histogram(smcInt, bins=np.arange(0, 17), weights=p[0])
        pmc = mpi.global_sum(jnp.array(pmc)[None,None,...])
        pmc = pmc / jnp.sum(pmc)

        self.assertTrue(jnp.max(jnp.abs(pmc - pex.reshape((-1,))[:16])) < 2e-3)

    def test_autoregressive_sampling_with_lstm(self):

        L = 4

        # Set up variational wave function
        rnn = nets.RNN1DGeneral(L=L, hiddenSize=5, cell="LSTM", realValuedParams=True, realValuedOutput=True, inputDim=2)
        rbm = nets.RBM(numHidden=2, bias=False)

        model = jVMC.nets.two_nets_wrapper.TwoNets((rnn, rbm))
        orbit = jVMC.util.symmetries.get_orbit_1D(L, "translation")
        net = nets.sym_wrapper.SymNet(net=model, orbit=orbit, avgFun=jVMC.nets.sym_wrapper.avgFun_Coefficients_Sep)
        psi = NQS(net)

        # Set up exact sampler
        exactSampler = sampler.ExactSampler(psi, L)

        # Set up MCMC sampler
        mcSampler = sampler.MCSampler(psi, (L,), random.PRNGKey(0), numChains=777)

        # Compute exact probabilities
        _, logPsi, pex = exactSampler.sample()
        pex = mpi.gather(pex)

        numSamples = 1000000
        smc, logPsi, p = mcSampler.sample(numSamples=numSamples)

        self.assertTrue(jnp.max(jnp.abs(jnp.real(psi(smc) - logPsi))) < 1e-12)

        smc = smc.reshape((smc.shape[0] * smc.shape[1], -1))

        self.assertTrue(mpi.global_sum(jnp.array([smc.shape[0],])[None,None,...]) >= numSamples)

        # Compute histogram of sampled configurations
        smcInt = jax.vmap(state_to_int)(smc)
        pmc, _ = np.histogram(smcInt, bins=np.arange(0, 17), weights=p[0])
        pmc = mpi.global_sum(jnp.array(pmc)[None,None,...])
        pmc = pmc / jnp.sum(pmc)

        self.assertTrue(jnp.max(jnp.abs(pmc - pex.reshape((-1,))[:16])) < 1e-3)

    def test_autoregressive_sampling_with_gru(self):

        L = 4

        # Set up variational wave function
        rnn = nets.RNN1DGeneral(L=L, hiddenSize=5, cell="GRU", realValuedParams=True, realValuedOutput=True, inputDim=2)
        rbm = nets.RBM(numHidden=2, bias=False)

        psi = NQS((rnn,rbm))

        # Set up exact sampler
        exactSampler = sampler.ExactSampler(psi, L)

        # Set up MCMC sampler
        mcSampler = sampler.MCSampler(psi, (L,), random.PRNGKey(0), numChains=777)

        # Compute exact probabilities
        _, logPsi, pex = exactSampler.sample()
        pex = mpi.gather(pex)

        numSamples = 1000000
        smc, logPsi, p = mcSampler.sample(numSamples=numSamples)

        self.assertTrue(jnp.max(jnp.abs(jnp.real(psi(smc) - logPsi))) < 1e-12)

        smc = smc.reshape((smc.shape[0] * smc.shape[1], -1))

        self.assertTrue(mpi.global_sum(jnp.array([smc.shape[0],])[None,None,...]) >= numSamples)

        # Compute histogram of sampled configurations
        smcInt = jax.vmap(state_to_int)(smc)
        pmc, _ = np.histogram(smcInt, bins=np.arange(0, 17), weights=p[0])
        pmc = mpi.global_sum(jnp.array(pmc)[None,None,...])
        pmc = pmc / jnp.sum(pmc)

        self.assertTrue(jnp.max(jnp.abs(pmc - pex.reshape((-1,))[:16])) < 1e-3)

    def test_autoregressive_sampling_with_rnn2d(self):

        L = 2

        # Set up variational wave function
        #rnn = nets.RNN2D( L=L, hiddenSize=5 )
        rnn = nets.RNN2DGeneral(L=L, hiddenSize=5, cell="RNN", realValuedParams=True, realValuedOutput=True)

        orbit = jVMC.util.symmetries.get_orbit_2D_square(L)
        psi = NQS((rnn,rnn), orbit=orbit, avgFun=jVMC.nets.sym_wrapper.avgFun_Coefficients_Sep)

        # Set up exact sampler
        exactSampler = sampler.ExactSampler(psi, (L, L))

        # Set up MCMC sampler
        mcSampler = sampler.MCSampler(psi, (L, L), random.PRNGKey(0), numChains=777)

        # Compute exact probabilities
        _, logPsi, pex = exactSampler.sample()
        pex = mpi.gather(pex)

        self.assertTrue(jnp.abs(jnp.sum(pex) - 1.) < 1e-12)

        numSamples = 1000000
        smc, p, _ = mcSampler.sample(numSamples=numSamples)

        self.assertTrue(jnp.max(jnp.abs(jnp.real(psi(smc) - p))) < 1e-12)

        smc = smc.reshape((smc.shape[0] * smc.shape[1], -1))

        self.assertTrue(mpi.global_sum(jnp.array([smc.shape[0],])[None,None,...]) >= numSamples)

        # Compute histogram of sampled configurations
        smcInt = jax.vmap(state_to_int)(smc)
        pmc, _ = np.histogram(smcInt, bins=np.arange(0, 17))
        pmc = mpi.global_sum(jnp.array(pmc)[None,None,...])
        pmc = pmc / jnp.sum(pmc)

        self.assertTrue(jnp.max(jnp.abs(pmc - pex.reshape((-1,))[:16])) < 1e-3)

    def test_autoregressive_sampling_with_rnn2d_symmetric(self):

        L = 2

        # Set up variational wave function
        #rnn = nets.RNN2D( L=L, hiddenSize=5 )
        rnn = nets.RNN2DGeneral(L=L, hiddenSize=5, cell="RNN", realValuedParams=True, realValuedOutput=True)

        symms = {"rotation": {"use": False, "factor": 1},
                 "translation": {"use": False, "factor": 1},
                 "reflection": {"use": False, "factor": 1},
                 "spinflip": {"use": False, "factor": 1},
                 }
        orbit = jVMC.util.symmetries.get_orbit_2D_square(L, "translation")
        psi = NQS((rnn,rnn), orbit=orbit, avgFun=jVMC.nets.sym_wrapper.avgFun_Coefficients_Sep)

        # Set up exact sampler
        exactSampler = sampler.ExactSampler(psi, (L, L))

        # Set up MCMC sampler
        mcSampler = sampler.MCSampler(psi, (L, L), random.PRNGKey(0), numChains=777)

        # Compute exact probabilities
        _, logPsi, pex = exactSampler.sample()
        pex = mpi.gather(pex)

        self.assertTrue(jnp.abs(jnp.sum(pex) - 1.) < 1e-12)

        numSamples = 1000000
        smc, logPsi, p = mcSampler.sample(numSamples=numSamples)

        self.assertTrue(jnp.max(jnp.abs(jnp.real(psi(smc) - logPsi))) < 1e-12)

        smc = smc.reshape((smc.shape[0] * smc.shape[1], -1))

        self.assertTrue(mpi.global_sum(jnp.array([smc.shape[0],])[None,None,...]) >= numSamples)

        # Compute histogram of sampled configurations
        smcInt = jax.vmap(state_to_int)(smc)
        pmc, _ = np.histogram(smcInt, bins=np.arange(0, 17), weights=p[0])
        pmc = mpi.global_sum(jnp.array(pmc)[None,None,...])
        pmc = pmc / jnp.sum(pmc)

        self.assertTrue(jnp.max(jnp.abs(pmc - pex.reshape((-1,))[:16])) < 1e-3)

    def test_autoregressive_sampling_with_lstm2d(self):

        L = 2

        # Set up variational wave function
        rnn = nets.RNN2DGeneral(L=L, hiddenSize=5, cell="LSTM", realValuedParams=True, realValuedOutput=True)
        psi = NQS((rnn,rnn))

        # Set up exact sampler
        exactSampler = sampler.ExactSampler(psi, (L, L))

        # Set up MCMC sampler
        mcSampler = sampler.MCSampler(psi, (L, L), random.PRNGKey(0), numChains=777)

        # Compute exact probabilities
        _, logPsi, pex = exactSampler.sample()
        pex = mpi.gather(pex)

        self.assertTrue(jnp.abs(jnp.sum(pex) - 1.) < 1e-12)

        numSamples = 1000000
        smc, logPsi, p = mcSampler.sample(numSamples=numSamples)

        self.assertTrue(jnp.max(jnp.abs(jnp.real(psi(smc) - logPsi))) < 1e-12)

        smc = smc.reshape((smc.shape[0] * smc.shape[1], -1))

        self.assertTrue(mpi.global_sum(jnp.array([smc.shape[0],])[None,None,...]) >= numSamples)

        # Compute histogram of sampled configurations
        smcInt = jax.vmap(state_to_int)(smc)
        pmc, _ = np.histogram(smcInt, bins=np.arange(0, 17), weights=p[0])
        pmc = mpi.global_sum(jnp.array(pmc)[None,None,...])
        pmc = pmc / jnp.sum(pmc)

        self.assertTrue(jnp.max(jnp.abs(pmc - pex.reshape((-1,))[:16])) < 1e-3)


class TestExactSampler(unittest.TestCase):

    def test_exact_sampler(self):
        L = 4

        weights = jnp.array(
            [0.23898957, 0.12614753, 0.19479055, 0.17325271, 0.14619853, 0.21392751,
             0.19648707, 0.17103704, -0.15457255, 0.10954413, 0.13228065, -0.14935214,
             -0.09963073, 0.17610707, 0.13386381, -0.14836467]
        )

        # Set up variational wave function
        rbm = nets.CpxRBM(numHidden=2, bias=False)
        psi = NQS(rbm)

        # Set up exact sampler
        exactSampler = sampler.ExactSampler(psi, L)

        p0 = psi.get_parameters()
        psi.set_parameters(weights)

        # Compute exact probabilities
        s, psi_s, pex = exactSampler.sample()

        import flax
        self.assertTrue(jnp.max((psi(s) - psi_s) / psi_s) < 1e-14)
        
        s, psi_s, pex = exactSampler.sample(parameters=p0)

        psi.set_parameters(p0)
        self.assertTrue(jnp.max((psi(s) - psi_s) / psi_s) < 1e-14)


if __name__ == "__main__":
    unittest.main()



import unittest

import jVMC
import jVMC.nets as nets
from jVMC.vqs import NQS

import jVMC.global_defs as global_defs

import jax
import jax.numpy as jnp
import numpy as np
from math import isclose

import flax.linen as nn
import jVMC.nets.activation_functions as act_funs


def get_shape(shape):
    return (global_defs.myDeviceCount,) + shape


class CpxRBM_nonHolomorphic(nn.Module):
    """Restricted Boltzmann machine with complex parameters.

    Initialization arguments:
        * ``s``: Computational basis configuration.
        * ``numHidden``: Number of hidden units.
        * ``bias``: ``Boolean`` indicating whether to use bias.

    """
    numHidden: int = 2
    bias: bool = False

    @nn.compact
    def __call__(self, s):

        layer = nn.Dense(self.numHidden, use_bias=self.bias,
                         **jVMC.nets.initializers.init_fn_args(kernel_init=jVMC.nets.initializers.cplx_init,
                                        bias_init=jax.nn.initializers.zeros,
                                        dtype=global_defs.tCpx)
                         )

        out = layer(2 * s.ravel() - 1)
        out = out + jnp.real(out) * 1e-2
        return jnp.sum(act_funs.log_cosh(out))

# ** end class CpxRBM_nonHolomorphic

class Simple_nonHolomorphic(nn.Module):
    """Restricted Boltzmann machine with complex parameters.

    Initialization arguments:
        * ``s``: Computational basis configuration.
        * ``numHidden``: Number of hidden units.
        * ``bias``: ``Boolean`` indicating whether to use bias.

    """
    numHidden: int = 2
    bias: bool = False

    @nn.compact
    def __call__(self, s):

        z = self.param('z', jVMC.nets.initializers.cplx_init, (1,), global_defs.tCpx)
    
        return jnp.sum(jnp.conj(z))

# ** end class Simple_nonHolomorphic

class MatrixMultiplication_NonHolomorphic(nn.Module):
    holo: bool = False

    @nn.compact
    def __call__(self, s):
        layer1 = nn.Dense(1, use_bias=False, **jVMC.nets.initializers.init_fn_args(dtype=global_defs.tCpx))
        out = layer1(2 * s.ravel() - 1)
        if not self.holo:
            out = out + 1e-1 * jnp.real(out)
        return jnp.sum(out)
    
# ** end class MatrixMultiplication_NonHolomorphic


class TestGradients(unittest.TestCase):

    def test_automatic_holomorphicity_recognition(self):

        for k in range(10):
            L = 3
            rbm = nets.CpxRBM(numHidden=2**k, bias=True)
            orbit = jVMC.util.symmetries.get_orbit_1D(L)
            net = nets.sym_wrapper.SymNet(net=rbm, orbit=orbit)
            s = jnp.zeros(get_shape((4, 3)), dtype=np.int32)
            psiC = NQS(net)
            psiC(s)

            self.assertTrue(psiC.holomorphic)

    def test_gradients_cpx(self):

        dlist = [jax.devices()[0], jax.devices()]

        for ds in dlist:

            global_defs.set_pmap_devices(ds)

            L = 3
            rbm = nets.CpxRBM(numHidden=2, bias=True)

            orbit = jVMC.util.symmetries.get_orbit_1D(L)
            net = nets.sym_wrapper.SymNet(net=rbm, orbit=orbit)
            psiC = NQS(net)

            s = jnp.zeros(get_shape((4, 3)), dtype=np.int32)
            s = s.at[..., 0, 1].set(1)
            s = s.at[..., 2, 2].set(1)

            psi0 = psiC(s)
            G = psiC.gradients(s)
            delta = 1e-6
            params = psiC.get_parameters()
            for j in range(G.shape[-1]):
                u = jnp.zeros(G.shape[-1], dtype=global_defs.tReal).at[j].set(1)
                psiC.update_parameters(delta * u)
                psi1 = psiC(s)
                psiC.set_parameters(params)

                # Finite difference gradients
                Gfd = (psi1 - psi0) / delta
                with self.subTest(i=j):
                    self.assertTrue(jnp.max(jnp.abs(Gfd - G[..., j])) < 1e-2)

    def test_gradients_real(self):

        dlist = [jax.devices()[0], jax.devices()]

        for ds in dlist:

            global_defs.set_pmap_devices(ds)

            L = 3
            rbmModel1 = nets.RBM(numHidden=2, bias=True)
            rbmModel2 = nets.RBM(numHidden=3, bias=True)
            psi = NQS((rbmModel1, rbmModel2))

            s = jnp.zeros(get_shape((4, 3)), dtype=np.int32)
            s = s.at[..., 0, 1].set(1)
            s = s.at[..., 2, 2].set(1)

            psi0 = psi(s)
            G = psi.gradients(s)
            delta = 1e-5
            params = psi.get_parameters()
            for j in range(G.shape[-1]):
                u = jnp.zeros(G.shape[-1], dtype=jVMC.global_defs.tReal).at[j].set(1)
                psi.update_parameters(delta * u)
                psi1 = psi(s)
                psi.set_parameters(params)

                # Finite difference gradients
                Gfd = (psi1 - psi0) / delta

                with self.subTest(i=j):
                    self.assertTrue(jnp.max(jnp.abs(Gfd - G[..., j])) < 1e-2)

    def test_gradients_nonholomorphic(self):

        dlist = [jax.devices()[0], jax.devices()]

        for ds in dlist:

            global_defs.set_pmap_devices(ds)

            L = 3
            model = nets.RNN1DGeneral(L=L)
            psi = NQS(model)

            s = jnp.zeros(get_shape((4, 3)), dtype=np.int32)
            s = s.at[..., 0, 1].set(1)
            s = s.at[..., 2, 2].set(1)

            psi0 = psi(s)
            G = psi.gradients(s)
            delta = 1e-5
            params = psi.get_parameters()
            for j in range(G.shape[-1]):
                u = jnp.zeros(G.shape[-1], dtype=jVMC.global_defs.tReal).at[j].set(1)
                psi.update_parameters(delta * u)
                psi1 = psi(s)
                psi.set_parameters(params)

                # Finite difference gradients
                Gfd = (psi1 - psi0) / delta

                with self.subTest(i=j):
                    self.assertTrue(jnp.max(jnp.abs(Gfd - G[..., j])) < 1e-2)

    
    def test_gradients_complex_nonholomorphic(self):
        
        dlist=[jax.devices()[0], jax.devices()]

        for ds in dlist:

            global_defs.set_pmap_devices(ds)

            model = Simple_nonHolomorphic()

            s=jnp.zeros(get_shape((4,3)),dtype=np.int32)
            s=s.at[...,0,1].set(1)
            s=s.at[...,2,2].set(1)
            
            psi = NQS(model)
            psi0 = psi(s)
            G = psi.gradients(s)
            delta=1e-5
            params = psi.get_parameters()
            for j in range(G.shape[-1]):
                u = jnp.zeros(G.shape[-1], dtype=jVMC.global_defs.tCpx).at[j].set(1)
                psi.update_parameters(delta * u)
                psi1 = psi(s)
                psi.set_parameters(params)

                # Finite difference gradients
                Gfd = (psi1-psi0) / delta

                with self.subTest(i=j):
                    self.assertTrue( jnp.max( jnp.abs( Gfd - G[...,j] ) ) < 1e-2 )
        
        for ds in dlist:

            global_defs.set_pmap_devices(ds)

            model = CpxRBM_nonHolomorphic()

            s=jnp.zeros(get_shape((4,3)),dtype=np.int32)
            s=s.at[...,0,1].set(1)
            s=s.at[...,2,2].set(1)
            
            psi = NQS(model)
            psi0 = psi(s)
            G = psi.gradients(s)
            delta=1e-5
            params = psi.get_parameters()
            for j in range(G.shape[-1]):
                u = jnp.zeros(G.shape[-1], dtype=jVMC.global_defs.tCpx).at[j].set(1)
                psi.update_parameters(delta * u)
                psi1 = psi(s)
                psi.set_parameters(params)

                # Finite difference gradients
                Gfd = (psi1-psi0) / delta

                with self.subTest(i=j):
                    self.assertTrue( jnp.max( jnp.abs( Gfd - G[...,j] ) ) < 1e-2 )

        for ds in dlist:

            global_defs.set_pmap_devices(ds)

            model = MatrixMultiplication_NonHolomorphic(holo=False)

            s=jnp.zeros(get_shape((1,4)),dtype=np.int32)
            
            psi = NQS(model)
            psi0 = psi(s)
            G = psi.gradients(s)
            ref = jnp.array([-1.1+0.j, -1.1+0.j, -1.1+0.j, -1.1+0.j, -0.-1.j, -0.-1.j, -0.-1.j, -0.-1.j])
            self.assertTrue( jnp.allclose( G.ravel(), ref ) )


    def test_gradient_dict(self):
        
        dlist=[[jax.devices()[0],], jax.devices()]

        for ds in dlist:

            global_defs.set_pmap_devices(ds)

            net = jVMC.nets.CpxRBM(numHidden=8, bias=False)
            psi = jVMC.vqs.NQS(net, seed=1234)  # Variational wave function

            s = jnp.zeros((len(ds),3,4))
            psi(s)

            g1 = psi.gradients(s)
            g2 = psi.gradients_dict(s)["Dense_0"]["kernel"]

            self.assertTrue(isclose(jnp.linalg.norm(g1-g2),0.0))

class TestEvaluation(unittest.TestCase):

    def test_evaluation_cpx(self):

        dlist = [jax.devices()[0], jax.devices()]

        for ds in dlist:

            global_defs.set_pmap_devices(ds)

            L = 3
            model = nets.CpxRBM(numHidden=2, bias=True)
            psiC = NQS(model)

            s = jnp.zeros(get_shape((4, L)), dtype=np.int32)
            s = s.at[..., 0, 1].set(1)
            s = s.at[..., 2, 2].set(1)

            cpxCoeffs = psiC(s)
            f, p = psiC.get_sampler_net()
            realCoeffs = global_defs.pmap_for_my_devices(lambda x: jax.vmap(lambda y: f(p, y))(x))(s)

            self.assertTrue(jnp.linalg.norm(jnp.real(cpxCoeffs) - realCoeffs) < 1e-6)


if __name__ == "__main__":
    unittest.main()



import unittest

import jVMC
import jVMC.operator as op
import jax
import jax.numpy as jnp


class TestPOVM(unittest.TestCase):
    def prepare_net(self, L, dt, hiddenSize=1, depth=1, cell="RNN"):
        def copy_dict(a):
            b = {}
            for key, value in a.items():
                if type(value) == type(a):
                    b[key] = copy_dict(value)
                else:
                    b[key] = value
            return b

        sample_shape = (L,)

        self.psi = jVMC.util.util.init_net({"batch_size": 200, "net1":
                                            {"type": "RNN",
                                             "translation": {"use": True, "factor": 1},
                                             "parameters": {"inputDim": 4,
                                                            "realValuedOutput": True,
                                                            "realValuedParams": True,
                                                            "logProbFactor": 1, "hiddenSize": hiddenSize, "L": L, "depth": depth, "cell": cell}}},
                                           sample_shape, 1234)

        system_data = {"dim": "1D", "L": L}
        self.povm = op.POVM(system_data)

        prob_dist = jVMC.operator.povm.get_1_particle_distributions("y_up", self.povm)
        prob_dist /= prob_dist[0]
        biases = jnp.log(prob_dist[1:])
        params = copy_dict(self.psi._param_unflatten(self.psi.get_parameters()))

        params["net"]["outputDense"]["bias"] = biases
        params["net"]["outputDense"]["kernel"] = 1e-15 * params["net"]["outputDense"]["kernel"]
        params = jnp.concatenate([p.ravel()
                                  for p in jax.tree_util.tree_flatten(params)[0]])
        self.psi.set_parameters(params)

        self.sampler = jVMC.sampler.ExactSampler(self.psi, (L,), lDim=4, logProbFactor=1)

        self.tdvpEquation = jVMC.util.tdvp.TDVP(self.sampler, rhsPrefactor=-1.,
                                                pinvTol=0.0, pinvCutoff=1e-6, diagonalShift=0, makeReal='real', crossValidation=False)

        #self.stepper = jVMC.util.stepper.Euler(timeStep=dt)  # ODE integrator
        self.stepper = jVMC.util.stepper.Heun(timeStep=dt)  # ODE integrator

    def test_matrix_to_povm(self):
        unity = jnp.eye(2, dtype=jVMC.global_defs.tCpx)
        zero_matrix = jnp.zeros((2, 2), dtype=jVMC.global_defs.tCpx)

        system_data = {"dim": "1D", "L": 2}
        povm = jVMC.operator.POVM(system_data)

        self.assertTrue(jnp.isclose(op.matrix_to_povm(unity, povm.M, povm.T_inv, mode='observable'),
                                    jnp.ones(4)).all())
        self.assertTrue(jnp.isclose(op.matrix_to_povm(unity, povm.M, povm.T_inv, mode='unitary'),
                                    jnp.zeros((4, 4))).all())
        self.assertTrue(jnp.isclose(op.matrix_to_povm(unity, povm.M, povm.T_inv, mode='dissipative'),
                                    jnp.zeros((4, 4))).all())
        self.assertTrue(jnp.isclose(op.matrix_to_povm(unity, povm.M, povm.T_inv, mode='imaginary'),
                                    -2*jnp.eye(4)).all())

        self.assertTrue(jnp.isclose(op.matrix_to_povm(zero_matrix, povm.M, povm.T_inv, mode='observable'),
                                    jnp.zeros(4)).all())
        self.assertTrue(jnp.isclose(op.matrix_to_povm(zero_matrix, povm.M, povm.T_inv, mode='unitary'),
                                    jnp.zeros((4, 4))).all())
        self.assertTrue(jnp.isclose(op.matrix_to_povm(zero_matrix, povm.M, povm.T_inv, mode='dissipative'),
                                    jnp.zeros((4, 4))).all())
        self.assertTrue(jnp.isclose(op.matrix_to_povm(zero_matrix, povm.M, povm.T_inv, mode='imaginary'),
                                    jnp.zeros((4, 4))).all())

        self.assertRaises(ValueError, op.matrix_to_povm, zero_matrix, povm.M, povm.T_inv, mode='wrong_mode')

    def test_adding_operator(self):
        unity = jnp.eye(2, dtype=jVMC.global_defs.tCpx)
        zero_matrix = jnp.zeros((2, 2), dtype=jVMC.global_defs.tCpx)

        system_data = {"dim": "1D", "L": 2}
        povm = jVMC.operator.POVM(system_data)

        unity_povm = op.matrix_to_povm(unity, povm.M, povm.T_inv, mode='unitary')
        zeros_povm = op.matrix_to_povm(zero_matrix, povm.M, povm.T_inv, mode='dissipative')

        self.assertFalse("unity" in povm.operators.keys())
        self.assertFalse("zero" in povm.operators.keys())

        povm.add_unitary("unity", unity_povm)
        povm.add_dissipator("zero", zeros_povm)

        self.assertTrue("unity" in povm.operators.keys())
        self.assertTrue("zero" in povm.operators.keys())

        self.assertRaises(ValueError, povm.add_unitary, "zero", op.matrix_to_povm(zero_matrix, povm.M,
                                                                                  povm.T_inv, mode='unitary'))
        self.assertRaises(ValueError, povm.add_dissipator, "unity", op.matrix_to_povm(unity, povm.M,
                                                                                      povm.T_inv, mode='dissipative'))

    def test_time_evolution_one_site(self):
        # This tests the time evolution of a sample system and compares it with the analytical solution

        L = 3
        Tmax = 0.5
        dt = 2E-3

        self.prepare_net(L, dt, hiddenSize=1, depth=1)

        Lindbladian = op.POVMOperator(self.povm)
        for l in range(L):
            Lindbladian.add({"name": "X", "strength": 3.0, "sites": (l,)})
            Lindbladian.add({"name": "dephasing", "strength": 1.0, "sites": (l,)})

        res = {"X": [], "Y": [], "Z": []}

        times = []
        t=0.
        while t<Tmax:

            result = jVMC.operator.povm.measure_povm(Lindbladian.povm, self.sampler)
            for dim in ["X", "Y", "Z"]:
                res[dim].append(result[dim]["mean"])
            times.append(t)
            if t>0.005:
                self.stepper.set_dt(3e-2)

            dp, stepSize = self.stepper.step(0, self.tdvpEquation, self.psi.get_parameters(), hamiltonian=Lindbladian,
                                      psi=self.psi)
            
            t += stepSize

            self.psi.set_parameters(dp)

        times = jnp.array(times)

        # Analytical solution
        w = jnp.sqrt(35)
        Sx_avg = jnp.zeros_like(times)
        Sy_avg = (w * jnp.cos(w * times) - jnp.sin(w * times)) / w * jnp.exp(-times)
        Sz_avg = 6 / w * jnp.sin(w * times) * jnp.exp(-times)

        # print(Sz_avg-jnp.asarray(res["Z"]))
        self.assertTrue(jnp.allclose(Sx_avg, jnp.asarray(res["X"]), atol=1e-2))
        self.assertTrue(jnp.allclose(Sy_avg, jnp.asarray(res["Y"]), atol=1e-2))
        self.assertTrue(jnp.allclose(Sz_avg, jnp.asarray(res["Z"]), atol=1e-2))

    def test_time_evolution_two_site(self):
        # This tests the time evolution of a sample system and compares it with the analytical solution

        L = 3
        Tmax = 0.25
        dt = 2E-3

        self.prepare_net(L, dt, hiddenSize=3, depth=1)

        sx = op.get_paulis()[0]
        XX_ = jnp.kron(sx, sx)
        M_2_body = jnp.array(
            [[jnp.kron(self.povm.M[i], self.povm.M[j]) for j in range(4)] for i in range(4)]).reshape(16, 4, 4)
        T_inv_2_body = jnp.kron(self.povm.T_inv, self.povm.T_inv)

        self.povm.add_dissipator("XX_", op.matrix_to_povm(XX_, M_2_body, T_inv_2_body, mode="dissipative"))

        Lindbladian = op.POVMOperator(self.povm)
        Lindbladian.add({"name": "XX_", "strength": 1.0, "sites": (0, 1)})
        Lindbladian.add({"name": "Z", "strength": 3.0, "sites": (0,)})
        Lindbladian.add({"name": "Z", "strength": 3.0, "sites": (1,)})
        Lindbladian.add({"name": "Z", "strength": 3.0, "sites": (2,)})

        res = {"X": [], "Y": [], "Z": []}

        times = []
        t=0.
        while t<Tmax:

            result = jVMC.operator.povm.measure_povm(Lindbladian.povm, self.sampler)
            for dim in ["X", "Y", "Z"]:
                res[dim].append(result[dim]["mean"])
            times.append(t)
            if t>0.005:
                self.stepper.set_dt(2.5e-2)

            dp, stepSize = self.stepper.step(0, self.tdvpEquation, self.psi.get_parameters(), hamiltonian=Lindbladian,
                                      psi=self.psi)
            
            t += stepSize

            self.psi.set_parameters(dp)

        times = jnp.array(times)

        # Analytical solution
        w = jnp.sqrt(35)
        Sx_avg = -jnp.sin(6 * times) / 3 - 4 / w * jnp.sin(w * times) * jnp.exp(-times)
        Sy_avg = jnp.cos(6 * times) / 3 + (2 / 3 * jnp.cos(w * times) - 2 / 3 / w * jnp.sin(w * times)) * jnp.exp(-times)
        Sz_avg = jnp.zeros_like(times)

        self.assertTrue(jnp.allclose(Sx_avg, jnp.asarray(res["X"]), atol=1e-2))
        self.assertTrue(jnp.allclose(Sy_avg, jnp.asarray(res["Y"]), atol=1e-2))
        self.assertTrue(jnp.allclose(Sz_avg, jnp.asarray(res["Z"]), atol=1e-2))

    def test_time_evolution_three_site(self):
        # This tests the time evolution of a sample system and compares it with the analytical solution

        L = 3
        Tmax = 0.2
        dt = 5E-4

        self.prepare_net(L, dt, hiddenSize=3, depth=1)

        sx = op.get_paulis()[0]
        XXX = jnp.kron(jnp.kron(sx, sx), sx)
        M_3_body = jnp.array(
            [[[jnp.kron(jnp.kron(self.povm.M[i], self.povm.M[j]), self.povm.M[k]) for j in range(4)] for i in range(4)]
             for k in range(4)]).reshape(64, 8, 8)
        T_inv_3_body = jnp.kron(jnp.kron(self.povm.T_inv, self.povm.T_inv), self.povm.T_inv)

        self.povm.add_dissipator("XXX", op.matrix_to_povm(XXX, M_3_body, T_inv_3_body, mode="dissipative"))

        Lindbladian = op.POVMOperator(self.povm)
        Lindbladian.add({"name": "XXX", "strength": 1.0, "sites": (0, 1, 2)})
        Lindbladian.add({"name": "Z", "strength": 3.0, "sites": (0,)})
        Lindbladian.add({"name": "Z", "strength": 3.0, "sites": (1,)})
        Lindbladian.add({"name": "Z", "strength": 3.0, "sites": (2,)})

        res = {"X": [], "Y": [], "Z": []}

        times = []
        t=0.
        while t<Tmax:

            result = jVMC.operator.povm.measure_povm(Lindbladian.povm, self.sampler)
            for dim in ["X", "Y", "Z"]:
                res[dim].append(result[dim]["mean"])
            times.append(t)
            if t>0.003:
                self.stepper.set_dt(1e-2)

            dp, stepSize = self.stepper.step(0, self.tdvpEquation, self.psi.get_parameters(), hamiltonian=Lindbladian,
                                      psi=self.psi)
            
            t += stepSize

            self.psi.set_parameters(dp)

        times = jnp.array(times)

        # Analytical solution
        w = jnp.sqrt(35)
        Sx_avg = -6 * jnp.sin(w * times) * jnp.exp(-times) / w
        Sy_avg = jnp.cos(w * times) * jnp.exp(-times) - jnp.sin(w * times) * jnp.exp(-times) / w
        Sz_avg = jnp.zeros_like(times)

        print(Sx_avg - jnp.asarray(res["X"]))
        self.assertTrue(jnp.allclose(Sx_avg, jnp.asarray(res["X"]), atol=1e-2))
        self.assertTrue(jnp.allclose(Sy_avg, jnp.asarray(res["Y"]), atol=1e-2))
        self.assertTrue(jnp.allclose(Sz_avg, jnp.asarray(res["Z"]), atol=1e-2))

    def test_ground_state_search(self):
        L = 2
        dt = 1E-2

        self.prepare_net(L, dt)

        sz = op.get_paulis()[2]
        self.povm.add_imaginary("imag_Z", op.matrix_to_povm(sz, self.povm.M, self.povm.T_inv, mode='imag'))

        Lindbladian = op.POVMOperator(self.povm)
        Lindbladian.add({"name": "imag_Z", "strength": 4., "sites": (0, )})
        Lindbladian.add({"name": "imag_Z", "strength": 4., "sites": (1, )})

        def measure_energy(confs, probs):
            return jnp.sum(jVMC.mpi_wrapper.global_mean(self.povm.observables["Z"][confs], probs))

        for i in range(40):
            dp, _ = self.stepper.step(0, self.tdvpEquation, self.psi.get_parameters(), hamiltonian=Lindbladian,
                                      psi=self.psi)
            self.psi.set_parameters(dp)

        confs, _, probs = self.sampler.sample()
        self.assertTrue(jnp.allclose(measure_energy(confs, probs), -2, atol=1e-2))


if __name__ == '__main__':
    unittest.main()



import unittest

import jax
jax.config.update("jax_enable_x64", True)

import jax.random as random
import jax.numpy as jnp

import numpy as np

import jVMC
import jVMC.operator as op
import jVMC.sampler
import jVMC.nets as nets
from jVMC.vqs import NQS
import jVMC.global_defs as global_defs

import flax.linen as nn
class Target(nn.Module):
  """Target wave function, returns a vector with the same dimension as the Hilbert space

    Initialization arguments:
        * ``L``: System size
        * ``d``: local Hilbert space dimension
        * ``delta``: small number to avoid log(0)

    """
  L: int
  d: float = 2.00
  delta: float = 1e-15

  @nn.compact
  def __call__(self, s):
    kernel = self.param('kernel',
                        nn.initializers.constant(1),
                        (int(self.d**self.L)))
    # return amplitude for state s
    idx = ((self.d**jnp.arange(self.L)).dot(s[::-1])).astype(int) # NOTE that the state is reversed to account for different bit conventions used in openfermion
    return jnp.log(abs(kernel[idx]+self.delta)) + 1.j*jnp.angle(kernel[idx]) 


def get_shape(shape):
    return (global_defs.device_count(),) + shape


class TestOperator(unittest.TestCase):

    def test_nonzeros(self):

        L = 4
        lDim = 2
        key = random.PRNGKey(3)
        s = random.randint(key, (24, L), 0, 2, dtype=np.int32).reshape(get_shape((-1, L)))

        h = op.BranchFreeOperator()

        h += 2. * op.Sp(0)
        h += 2. * op.Sp(1)
        h += 2. * op.Sp(2)

        sp, matEl = h.get_s_primes(s)

        logPsi = jnp.ones(s.shape[:-1])
        logPsiSP = jnp.ones(sp.shape[:-1])

        tmp = h.get_O_loc_unbatched(logPsi, logPsiSP)

        self.assertTrue(jnp.sum(jnp.abs(tmp - 2. * jnp.sum(-(s[..., :3] - 1), axis=-1))) < 1e-7)

    def test_op_with_arguments(self):

        L = 4
        key = random.PRNGKey(3)
        s = random.randint(key, (24, L), 0, 2, dtype=np.int32).reshape(get_shape((-1, L)))

        h = op.BranchFreeOperator()

        def f(t):
            return 2.0 * t

        h.add(op.scal_opstr(f, (op.Sp(0),)))
        h.add(op.scal_opstr(f, (op.Sp(1),)))
        h.add(op.scal_opstr(f, (op.Sp(2),)))

        for t in [0.5, 2, 13.9]:
            sp, matEl = h.get_s_primes(s, t)

            logPsi = jnp.ones(s.shape[:-1])
            logPsiSP = jnp.ones(sp.shape[:-1])

            tmp = h.get_O_loc_unbatched(logPsi, logPsiSP)

            self.assertTrue(jnp.sum(jnp.abs(tmp - f(t) * jnp.sum(-(s[..., :3] - 1), axis=-1))) < 1e-7)

    def test_op_2d(self):

        L = 4
        key = random.PRNGKey(3)
        s = random.randint(key, (24, L, L), 0, 2, dtype=np.int32).reshape(get_shape((-1, L, L)))

        h = op.BranchFreeOperator()

        h += 0.3 * op.Sp(0)
        h += 1.1 * op.Sp(1)
        h += 0.15 * op.Sp(4)

        sp, matEl = h.get_s_primes(s)

        self.assertTrue(sp.shape[2:] == (L,L))


    def test_batched_Oloc(self):

        L = 4

        h = op.BranchFreeOperator()
        for i in range(L):
            h.add(op.scal_opstr(2., (op.Sx(i),)))
            h.add(op.scal_opstr(2., (op.Sy(i), op.Sz((i + 1) % L))))

        rbm = nets.CpxRBM(numHidden=2, bias=False)
        psi = NQS(rbm)

        mcSampler = jVMC.sampler.MCSampler(psi, (L,), random.PRNGKey(0), updateProposer=jVMC.sampler.propose_spin_flip, numChains=1)

        numSamples = 100
        s, logPsi, _ = mcSampler.sample(numSamples=numSamples)

        sp, matEl = h.get_s_primes(s)
        logPsiSp = psi(sp)
        Oloc1 = h.get_O_loc_unbatched(logPsi, logPsiSp)

        batchSize = 13
        Oloc2 = h.get_O_loc_batched(s, psi, logPsi, batchSize)

        self.assertTrue(jnp.abs(jnp.sum(Oloc1) - jnp.sum(Oloc2)) < 1e-5)

    def test_batched_Oloc2(self):
        L = 4

        hamilton_unbatched = op.BranchFreeOperator()
        hamilton_batched = op.BranchFreeOperator(ElocBatchSize=13)
        for i in range(L):
            hamilton_unbatched.add(op.scal_opstr(2., (op.Sx(i),)))
            hamilton_unbatched.add(op.scal_opstr(2., (op.Sy(i), op.Sz((i + 1) % L))))
            hamilton_batched.add(op.scal_opstr(2., (op.Sx(i),)))
            hamilton_batched.add(op.scal_opstr(2., (op.Sy(i), op.Sz((i + 1) % L))))

        rbm = nets.CpxRBM(numHidden=2, bias=False)
        psi = NQS(rbm)

        mcSampler = jVMC.sampler.MCSampler(psi, (L,), random.PRNGKey(0), updateProposer=jVMC.sampler.propose_spin_flip,
                                           numChains=1)

        numSamples = 100
        s, logPsi, _ = mcSampler.sample(numSamples=numSamples)

        Oloc1 = hamilton_unbatched.get_O_loc(s, psi, logPsi)

        Oloc2 = hamilton_batched.get_O_loc(s, psi, logPsi)

        self.assertTrue(jnp.abs(jnp.sum(Oloc1) - jnp.sum(Oloc2)) < 1e-5)

    def test_td_prefactor(self):

        hamiltonian = op.BranchFreeOperator()
        hamiltonian.add((op.Sz(0),))
        hamiltonian.add((op.Sz(1),))
        # hamiltonian.add((op.Sx(0),op.Sx(1)))
        hamiltonian.add(op.scal_opstr(0.1, (op.Sx(0), op.Sx(1))))

        hamiltonian.compile()

    def test_fermionic_operators(self):
        L = 2

        rbm = nets.CpxRBM(numHidden=2, bias=True)
        psi = NQS(rbm)

        sampler = jVMC.sampler.ExactSampler(psi, (L,))

        def commutator(i,j):
            Comm = op.BranchFreeOperator()
            # Comm.add(op.scal_opstr( 1., (op.annihilation(i), op.creation(j), ) ) )
            # Comm.add(op.scal_opstr( 1., (op.creation(j), op.annihilation(i), ) ) )
            Comm.add(op.scal_opstr( 1., (op.creation(j), op.annihilation(i)) ) )
            Comm.add(op.scal_opstr( 1., (op.annihilation(i), op.creation(j)) ) )
            return Comm

        observalbes_dict = {
                            "same_site": [commutator(0,0),commutator(1,1)], 
                            "distinct_site": [commutator(0,1),commutator(1,0)]
                            }
        out_dict = jVMC.util.util.measure(observalbes_dict, psi, sampler)

        self.assertTrue(
            jnp.allclose(
                jnp.concatenate(
                    (out_dict["same_site"]['mean'],
                     out_dict["distinct_site"]['mean'])),
                        jnp.array([1.,1.,0.,0.]),
                        rtol=1e-15)
            )
        
        self.assertTrue(
            jnp.allclose(
                jnp.concatenate(
                    (out_dict["same_site"]['variance'],
                     out_dict["distinct_site"]['variance'])),
                        jnp.array([0.,0.,0.,0.]),
                        rtol=1e-15)
            )
        

        t = - 1.0 # hopping
        mu = -2.0 # chemical potential
        V = 4.0 # interaction
        L = 4   # number of sites
        flavour = 2 # number of flavours
        flavourL = flavour*L # number of spins times sites

        # initalize the Hamitonian
        hamiltonian = op.BranchFreeOperator()
        # impurity definitions
        site1UP = 0
        site1DO = flavourL-1#//flavour
        # loop over the 1d lattice
        for i in range(0,flavourL//flavour):
            # interaction
            hamiltonian.add(op.scal_opstr( V, ( op.number(site1UP + i) , op.number(site1DO - i) ) ) )
            # chemical potential
            hamiltonian.add(op.scal_opstr(mu , ( op.number(site1UP + i) ,) ) )
            hamiltonian.add(op.scal_opstr(mu , ( op.number(site1DO - i) ,) ) )
            if i == flavourL//flavour-1:
                continue
            # up chain hopping
            hamiltonian.add(op.scal_opstr( t, ( op.creation(site1UP + i + 1), op.annihilation(site1UP + i) ,  ) ) )
            hamiltonian.add(op.scal_opstr( t, ( op.creation(site1UP + i), op.annihilation(site1UP + i + 1) ,  ) ) )
            # down chain hopping
            hamiltonian.add(op.scal_opstr( t, ( op.creation(site1DO - i - 1), op.annihilation(site1DO - i) ,  ) ) )
            hamiltonian.add(op.scal_opstr( t, ( op.creation(site1DO - i), op.annihilation(site1DO - i - 1) ,  ) ) )

        b = np.loadtxt("tests/data_ref/fermion_ref.txt", dtype=np.complex128)
        chi_model = Target(L=flavourL, d=2)
        chi = NQS(chi_model)
        chi(jnp.array(jnp.ones((1, 1, flavourL))))
        chi.set_parameters(b)
        chiSampler = jVMC.sampler.ExactSampler(chi, (flavourL,))
        s, logPsi, p = chiSampler.sample()
        sPrime, _ = hamiltonian.get_s_primes(s)
        Oloc = hamiltonian.get_O_loc(s, chi, logPsi)
        Omean = jVMC.mpi_wrapper.global_mean(Oloc,p)

        self.assertTrue(jnp.allclose(Omean, -9.95314531))
        
    def test_opstr(self):
        op1 = op.Sz(3)
        op2 = op.Sx(5)

        opstr1 = 13. * op1 * op2
        opstr2 = 1.j * opstr1 * op1

        self.assertTrue(jnp.allclose(opstr2[0](), 13.j))
        for o in opstr2[1:]:
            self.assertTrue(isinstance(o, (op.LocalOp, dict)))


if __name__ == "__main__":
    unittest.main()



import jax
import jax.numpy as jnp
import flax.linen as nn

from typing import Sequence, Union


class TwoNets(nn.Module):
    nets: Sequence[callable]

    def __post_init__(self):

        if "sample" in dir(self.nets[0]):
            self.sample = self._sample_fun

        super().__post_init__()


    def __call__(self, s):

        return self.nets[0](s) + 1j * self.nets[1](s)


    def eval_real(self, s):
        
        return self.nets[0](s)
        
    
    def _sample_fun(self, *args):

        return self.nets[0].sample(*args)



from jVMC.nets.activation_functions import *
from jVMC.nets.rbm import *
from jVMC.nets.cnn import *
from jVMC.nets.ffn import *
from jVMC.nets.rnn1d_general import *
from jVMC.nets.rnn2d_general import *
from jVMC.nets.sym_wrapper import *
from jVMC.nets.two_nets_wrapper import *



import unittest

import jax
jax.config.update("jax_enable_x64", True)

import jax.random as random
import jax.numpy as jnp

import numpy as np
from scipy.interpolate import interp1d

import jVMC
import jVMC.global_defs as global_defs
import jVMC.util.stepper as jVMCstepper
import jVMC.nets as nets
from jVMC.vqs import NQS
import jVMC.operator as op
import jVMC.sampler as sampler
import jVMC.mpi_wrapper as mpi

from jVMC.util import measure, ground_state_search

from functools import partial


class TestGsSearch(unittest.TestCase):
    def test_gs_search_cpx(self):
        L = 4
        J = -1.0
        hxs = [-1.3, -0.3]

        exEs = [-6.10160339, -4.09296160]

        for hx, exE in zip(hxs, exEs):
            # Set up variational wave function
            rbm = nets.CpxRBM(numHidden=6, bias=False)
            psi = NQS(rbm)

            # Set up hamiltonian for ground state search
            hamiltonianGS = op.BranchFreeOperator()
            for l in range(L):
                hamiltonianGS.add(op.scal_opstr(J, (op.Sz(l), op.Sz((l + 1) % L))))
                hamiltonianGS.add(op.scal_opstr(hx, (op.Sx(l), )))

            # Set up exact sampler
            exactSampler = sampler.ExactSampler(psi, L)

            delta = 2
            tdvpEquation = jVMC.util.TDVP(exactSampler, snrTol=1, pinvTol=0.0, pinvCutoff=1e-8, rhsPrefactor=1., diagonalShift=delta, makeReal='real')

            # Perform ground state search to get initial state
            ground_state_search(psi, hamiltonianGS, tdvpEquation, exactSampler, numSteps=100, stepSize=5e-2)

            obs = measure({"energy": hamiltonianGS}, psi, exactSampler)

            self.assertTrue(jnp.max(jnp.abs((obs['energy']['mean'] - exE) / exE)) < 1e-3)


class TestTimeEvolution(unittest.TestCase):
    def test_time_evolution(self):
        L = 4
        J = -1.0
        hx = -0.3

        weights = jnp.array(
            [0.23898957, 0.12614753, 0.19479055, 0.17325271, 0.14619853, 0.21392751,
             0.19648707, 0.17103704, -0.15457255, 0.10954413, 0.13228065, -0.14935214,
             -0.09963073, 0.17610707, 0.13386381, -0.14836467]
        )

        global_defs.set_pmap_devices(jax.devices()[0])

        # Set up variational wave function
        rbm = nets.CpxRBM(numHidden=2, bias=False)
        psi = NQS(rbm)
        psi(jnp.array([[[1, 1, 1, 1]]]))
        psi.set_parameters(weights)

        # Set up hamiltonian for time evolution
        hamiltonian = op.BranchFreeOperator()
        for l in range(L):
            hamiltonian.add(op.scal_opstr(J, (op.Sz(l), op.Sz((l + 1) % L))))
            hamiltonian.add(op.scal_opstr(hx, (op.Sx(l), )))

        # Set up ZZ observable
        ZZ = op.BranchFreeOperator()
        for l in range(L):
            ZZ.add((op.Sz(l), op.Sz((l + 1) % L)))

        # Set up exact sampler
        exactSampler = sampler.ExactSampler(psi, L)

        # Set up adaptive time stepper
        stepper = jVMCstepper.AdaptiveHeun(timeStep=1e-3, tol=1e-5)

        tdvpEquation = jVMC.util.TDVP(exactSampler, snrTol=1, pinvTol=0.0, pinvCutoff=1e-8, rhsPrefactor=1.j, diagonalShift=0., makeReal='imag')

        t = 0
        obs = []
        times = []
        times.append(t)
        newMeas = measure({'E': hamiltonian, 'ZZ': ZZ}, psi, exactSampler)
        obs.append([newMeas['E']['mean'], newMeas['ZZ']['mean']])
        while t < 0.5:
            dp, dt = stepper.step(0, tdvpEquation, psi.get_parameters(), hamiltonian=hamiltonian, psi=psi, numSamples=0)
            psi.set_parameters(dp)
            t += dt
            times.append(t)
            newMeas = measure({'E': [(hamiltonian, t)], 'ZZ': ZZ}, psi, exactSampler)
            obs.append([newMeas['E']['mean'], newMeas['ZZ']['mean']])

        obs = np.array(jnp.asarray(obs))

        # Check energy conservation
        obs[:, 0] = np.abs((obs[:, 0] - obs[0, 0]) / obs[0, 0])
        self.assertTrue(np.max(obs[:, 0]) < 1e-3)

        # Check observable dynamics
        zz = interp1d(np.array(times), obs[:, 1, 0])
        refTimes = np.arange(0, 0.5, 0.05)
        netZZ = zz(refTimes)
        refZZ = np.array(
            [0.882762129306284, 0.8936168721790617, 0.9257753299594491, 0.9779836185039352, 1.0482156449061142,
             1.1337654450614298, 1.231369697427413, 1.337354107391303, 1.447796176316155, 1.558696104640795,
             1.666147269524912, 1.7664978782554912, 1.8564960156892512, 1.9334113379450693, 1.9951280521882777,
             2.0402054805651546, 2.067904337137255, 2.078178742959828, 2.071635856483114, 2.049466698269522, 2.049466698269522]
        )
        self.assertTrue(np.max(np.abs(netZZ - refZZ[:len(netZZ)])) < 1e-3)


class TestTimeEvolutionMCSampler(unittest.TestCase):
    def test_time_evolution(self):
        L = 4
        J = -1.0
        hx = -0.3

        weights = jnp.array(
            [0.23898957, 0.12614753, 0.19479055, 0.17325271, 0.14619853, 0.21392751,
             0.19648707, 0.17103704, -0.15457255, 0.10954413, 0.13228065, -0.14935214,
             -0.09963073, 0.17610707, 0.13386381, -0.14836467]
        )

        # Set up variational wave function
        rbm = nets.CpxRBM(numHidden=2, bias=False)
        psi = NQS(rbm, batchSize=5000)
        psi(jnp.array([[[1, 1, 1, 1]]]))
        psi.set_parameters(weights)

        # Set up hamiltonian for time evolution
        hamiltonian = op.BranchFreeOperator()
        for l in range(L):
            hamiltonian.add(op.scal_opstr(J, (op.Sz(l), op.Sz((l + 1) % L))))
            hamiltonian.add(op.scal_opstr(hx, (op.Sx(l), )))

        # Set up ZZ observable
        ZZ = op.BranchFreeOperator()
        for l in range(L):
            ZZ.add((op.Sz(l), op.Sz((l + 1) % L)))

        # Set up exact sampler
        MCsampler = sampler.MCSampler(psi, (L,), jax.random.PRNGKey(0), numSamples=50000, updateProposer=sampler.propose_spin_flip, mu=1, numChains=500)

        # Set up adaptive time stepper
        stepper = jVMCstepper.AdaptiveHeun(timeStep=1e-3, tol=1e-4)

        tdvpEquation = jVMC.util.TDVP(MCsampler, snrTol=1, pinvTol=1e-8, rhsPrefactor=1.j, diagonalShift=0., makeReal='imag', crossValidation=True)

        t = 0
        obs = []
        times = []
        times.append(t)
        newMeas = measure({'E': hamiltonian, 'ZZ': ZZ}, psi, MCsampler)
        obs.append([newMeas['E']['mean'], newMeas['ZZ']['mean']])
        while t < 0.2:
            dp, dt = stepper.step(0, tdvpEquation, psi.get_parameters(), hamiltonian=hamiltonian, psi=psi, numSamples=5000)
            psi.set_parameters(dp)
            t += dt
            times.append(t)
            newMeas = measure({'E': [(hamiltonian, t)], 'ZZ': ZZ}, psi, MCsampler)
            obs.append([newMeas['E']['mean'], newMeas['ZZ']['mean']])

        obs = np.array(jnp.asarray(obs))

        # Check energy conservation
        obs[:, 0] = np.abs((obs[:, 0] - obs[0, 0]) / obs[0, 0])
        self.assertTrue(np.max(obs[:, 0]) < 1e-1)

        # Check observable dynamics
        zz = interp1d(np.array(times), obs[:, 1, 0])
        refTimes = np.arange(0, 0.2, 0.05)
        netZZ = zz(refTimes)
        refZZ = np.array(
            [0.882762129306284, 0.8936168721790617, 0.9257753299594491, 0.9779836185039352, 1.0482156449061142,
             1.1337654450614298, 1.231369697427413, 1.337354107391303, 1.447796176316155, 1.558696104640795,
             1.666147269524912, 1.7664978782554912, 1.8564960156892512, 1.9334113379450693, 1.9951280521882777,
             2.0402054805651546, 2.067904337137255, 2.078178742959828, 2.071635856483114, 2.049466698269522, 2.049466698269522]
        )
        self.assertTrue(np.max(np.abs(netZZ - refZZ[:len(netZZ)])) < 2e-2)


class TestSNRConsistency(unittest.TestCase):
    def test_snr_consistency(self):
        L = 4
        J = -1.0
        hx = -0.3

        weights = jnp.array(
            [0.23898957, 0.12614753, 0.19479055, 0.17325271, 0.14619853, 0.21392751,
             0.19648707, 0.17103704, -0.15457255, 0.10954413, 0.13228065, -0.14935214,
             -0.09963073, 0.17610707, 0.13386381, -0.14836467]
        )

        # Set up variational wave function
        rbm = nets.CpxRBM(numHidden=2, bias=False)
        orbit = jVMC.util.symmetries.get_orbit_1D(L, "translation")
        net = nets.sym_wrapper.SymNet(net=rbm, orbit=orbit)
        psi = NQS(net, batchSize=5000)
        psi(jnp.array([[[1, 1, 1, 1]]]))
        psi.set_parameters(weights)

        # Set up hamiltonian for time evolution
        hamiltonian = op.BranchFreeOperator()
        for l in range(L):
            hamiltonian.add(op.scal_opstr(J, (op.Sz(l), op.Sz((l + 1) % L))))
            hamiltonian.add(op.scal_opstr(hx, (op.Sx(l), )))

        # Set up exact sampler
        sampler = jVMC.sampler.MCSampler(psi, (L,), jax.random.PRNGKey(0), numSamples=10, updateProposer=jVMC.sampler.propose_spin_flip, mu=2, numChains=500)

        # Get sample
        sampleConfigs, sampleLogPsi, p = sampler.sample()

        # Evaluate local energy
        Eloc_old = hamiltonian.get_O_loc(sampleConfigs, psi, sampleLogPsi, 0.0)
        Eloc = jVMC.stats.SampledObs( Eloc_old, p)

        # Evaluate gradients
        sampleGradients_old = psi.gradients(sampleConfigs)
        sampleGradients = jVMC.stats.SampledObs( sampleGradients_old, p)

        self.F0 = (-1.j) * sampleGradients.covar(Eloc).ravel() 
        S = jVMC.util.imagFun( sampleGradients.covar() )

        ev, V = jnp.linalg.eigh(S)

        # old version
        subtract_helper_Eloc = global_defs.pmap_for_my_devices(lambda x, y: x - y, in_axes=(0, None))
        subtract_helper_grad = global_defs.pmap_for_my_devices(lambda x, y: x - y, in_axes=(0, None))
        transform_EO = global_defs.pmap_for_my_devices(lambda eo, v: jnp.matmul(eo, jnp.conj(v)), in_axes=(0, None))
        get_EO = global_defs.pmap_for_my_devices(lambda f, p, Eloc, grad: -f * jnp.multiply((p * Eloc)[:, None], jnp.conj(grad)),
                                                      in_axes=(None, 0, 0, 0), static_broadcasted_argnums=(0))
        Eloc_old = subtract_helper_Eloc(Eloc_old, mpi.global_mean(Eloc_old, p))
        sampleGradients_old = subtract_helper_grad(sampleGradients_old, mpi.global_mean(sampleGradients_old, p))
        EOdata = get_EO(-1., p, Eloc_old, sampleGradients_old) * mpi.globNumSamples
        EOdata = transform_EO(jVMC.util.imagFun(EOdata), V)
        EOdata.block_until_ready()
        rhoVar_old = mpi.global_variance(EOdata, jnp.ones(EOdata.shape[:2]) / mpi.globNumSamples)

        EO = sampleGradients.covar_data(Eloc).transform(linearFun = jnp.transpose(jnp.conj(V)),
                                                        nonLinearFun=lambda x: jVMC.util.imagFun(x))
        rhoVar_new = EO.var().ravel()

        self.assertTrue( jnp.allclose(rhoVar_old, rhoVar_new) )



if __name__ == "__main__":
    unittest.main()



import unittest

import jVMC.util.stepper as st

import jax
import jax.random as random
import jax.numpy as jnp
import numpy as np


class TestIntegrationHeun(unittest.TestCase):

    # Test adaptive integrator with multi-dimensional linear ODE
    def test_integration(self):

        def f(y,t,**args):
            return args['mat'].dot(y)

        def norm(y):
            return jnp.real(jnp.conjugate(y).dot(y))

        N = 4

        np.random.seed(123)
        stepper = st.AdaptiveHeun()
        mat = jnp.array(np.random.rand(N,N) + 1.j * np.random.rand(N,N))

        y0 = jnp.array(np.random.rand(N))
        y = y0.copy()
        t=0
        diffs=[]
        for k in range(100):
            y, dt = stepper.step(t,f,y,normFunction=norm,mat=mat)
            t+=dt
            yExact = jax.scipy.linalg.expm(t * mat).dot(y0)
            diff = y - yExact
            diffs.append(norm(diff)/N)

        self.assertTrue( jnp.max(jnp.array(diffs)) < 1e-5 )


if __name__ == "__main__":
    unittest.main()



import jax
jax.config.update("jax_enable_x64", True)

import jax.numpy as jnp
import jax.random as random

import time
import numpy as np

import jVMC
import jVMC.util.stepper as jVMCstepper
import jVMC.mpi_wrapper as mpi
import jVMC.nets.activation_functions as act_funs
import jVMC.util.symmetries as sym

import collections


def get_iterable(x):
    if isinstance(x, collections.abc.Iterable):
        return x
    else:
        return (x,)


def init_net(descr, dims, seed=0):

    def get_activation_functions(actFuns):

        if type(actFuns) is list:
            return tuple([act_funs.activationFunctions[fn] for fn in actFuns])

        return act_funs.activationFunctions[actFuns]

    netTypes = {
        "RBM": jVMC.nets.RBM,
        "FFN": jVMC.nets.FFN,
        "CNN": jVMC.nets.CNN,
        "RNN": jVMC.nets.RNN1DGeneral,
        "RNN2D": jVMC.nets.RNN2DGeneral,
        "CpxRBM": jVMC.nets.CpxRBM,
        "CpxCNN": jVMC.nets.CpxCNN
    }

    def get_net(descr, dims):

        return netTypes[descr["type"]](**descr["parameters"])

    if "actFun" in descr["net1"]["parameters"]:
        descr["net1"]["parameters"]["actFun"] = get_activation_functions(descr["net1"]["parameters"]["actFun"])

    symms = tuple()
    for key in ["translation", "rotation", "reflection", "spinflip"]:
        if key in descr:
            symms = symms + (key,)
    symm_factors = {}
    for key in symms:
        fac_key = key + "_factor"
        symm_factors[fac_key] = descr[fac_key] if fac_key in descr else 1

    if len(dims) == 2:
        orbit = sym.get_orbit_2D_square(dims[0], *symms, **symm_factors)
    if len(dims) == 1:
        orbit = sym.get_orbit_1D(dims[0], *symms, **symm_factors)

    if not "net2" in descr:

        model = get_net(descr["net1"], dims)
        isGenerator = "sample" in dir(model)

    else:

        if "actFun" in descr["net2"]["parameters"]:

            descr["net2"]["parameters"]["actFun"] = get_activation_functions(descr["net2"]["parameters"]["actFun"])

        model1 = get_net(descr["net1"], dims)
        model2 = get_net(descr["net2"], dims)
        model = jVMC.nets.two_nets_wrapper.TwoNets(net1=model1, net2=model2)
        isGenerator = "sample" in dir(model1)

    avgFun = jVMC.nets.sym_wrapper.avgFun_Coefficients_Sep if isGenerator else jVMC.nets.sym_wrapper.avgFun_Coefficients_Exp
    model = jVMC.nets.sym_wrapper.SymNet(orbit=orbit, net=model, avgFun=avgFun)
    psi = jVMC.vqs.NQS(model, batchSize=descr["batch_size"], seed=seed)

    psi(jnp.zeros((jVMC.global_defs.device_count(), 1) + dims, dtype=np.int32))

    return psi


def measure(observables, psi, sampler, numSamples=None):
    ''' This function measures expectation values of a given set of operators given a pure state.

    Arguments:
        * ``observables``: Dictionary of the form with operator names as keys and (lists of) operators as values, e.g.:

            .. code-block:: python

                { "op_name_1": [operator1, operator2], "op_name_2": operator3 }

            If an operator takes additional positional arguments, a tuple of the operator together with the arguments
            has to be passed instead. For example, assuming that ``operator3`` from above requires additional arguments ``*args``:

            .. code-block:: python

                { "op_name_1": [operator1, operator2], "op_name_2": [(operator3, *args)] }

        * ``psi``: Variational wave function (instance of ``jVMC.vqs.NQS``)
        * ``sampler``: Instance of ``jVMC.sampler`` used for sampling.
        * ``numSamples``: Number of samples (optional)

    Returns:
        A dictionary holding expectation values, variances, and MC error estimates for each operator. E.g. for the
        exemplary operator input given in `Arguments`:

        .. code-block:: python

            { 
                "op_name_1": { "mean": [mean1, mean2],
                               "variance": [var1, var2],
                               "MC_error": [err1, err2] },
                "op_name_2": { "mean": [mean3],
                               "variance": [var3],
                               "MC_error": [err3] }
            }

    '''
    # Get sample
    sampleConfigs, sampleLogPsi, p = sampler.sample(numSamples=numSamples)

    result = {}

    for name, ops in observables.items():

        tmpMeans = []
        tmpVariances = []
        tmpErrors = []

        for op in get_iterable(ops):

            args = ()
            if isinstance(op, collections.abc.Iterable):
                args = tuple(op[1:])
                op = op[0]

            Oloc = op.get_O_loc(sampleConfigs, psi, sampleLogPsi, *args)

            tmpMeans.append(mpi.global_mean(Oloc[..., None], p)[0])
            tmpVariances.append(mpi.global_variance(Oloc[..., None], p)[0])
            tmpErrors.append(jnp.sqrt(tmpVariances[-1]) / jnp.sqrt(sampler.get_last_number_of_samples()))

        result[name] = {}
        result[name]["mean"] = jnp.real(jnp.array(tmpMeans))
        result[name]["variance"] = jnp.real(jnp.array(tmpVariances))
        result[name]["MC_error"] = jnp.real(jnp.array(tmpErrors))

    return result


def ground_state_search(psi, ham, tdvpEquation, sampler, numSteps=200, varianceTol=1e-10, stepSize=1e-2, observables=None, outp=None):
    ''' This function performs a ground state search by Stochastic Reconfiguration.

    Arguments:
        * ``psi``: Variational wave function (``jVMC.vqs.NQS``)
        * ``ham``: Hamiltonian operator
        * ``tdvpEquation``: An instance of ``jVMC.util.TDVP`` or ``jVMC.util.MinSR``
        * ``numSteps``: Maximal number of steps
        * ``varianceTol``: Stopping criterion
        * ``stepSize``: Update step size (learning rate)
        * ``observables``: Observables to be measured during ground state search
        * ``outp``: ``None`` or instance of ``jVMC.util.OutputManager``.

    '''

    if "diagonalShift" in dir(tdvpEquation):
        delta = tdvpEquation.diagonalShift

    stepper = jVMCstepper.Euler(timeStep=stepSize)

    n = 0
    if outp is not None:
        if observables is not None:
            obs = measure(observables, psi, sampler)
            outp.write_observables(n, **obs)

    varE = 1.0

    while n < numSteps and varE > varianceTol:

        tic = time.perf_counter()

        dp, _ = stepper.step(0, tdvpEquation, psi.get_parameters(), hamiltonian=ham, psi=psi, numSamples=None, outp=outp)
        psi.set_parameters(dp)
        n += 1

        varE = tdvpEquation.get_energy_variance()

        if outp is not None:
            if observables is not None:
                obs = measure(observables, psi, sampler)
                outp.write_observables(n, **obs)

        if "set_diagonal_shift" in dir(tdvpEquation):
            delta = 0.95 * delta
            tdvpEquation.set_diagonal_shift(delta)

        if outp is not None:
            outp.print(" STEP %d" % (n))
            outp.print("   Energy mean: %f" % (tdvpEquation.get_energy_mean()))
            outp.print("   Energy variance: %f" % (varE))
            outp.print_timings(indent="   ")
            outp.print("   == Time for step: %fs" % (time.perf_counter() - tic))



import jax
from jax import vmap
import jax.numpy as jnp
import numpy as np

import jVMC.global_defs as global_defs
from mpi4py import MPI
from . import Operator

import functools

opDtype = global_defs.tCpx

# Common operators


def Id(idx=0, lDim=2):
    """Returns an identity operator

    Args:

    * ``idx``: Index of the local Hilbert space.
    * ``lDim``: Dimension of local Hilbert space.

    Returns:
        Dictionary defining an identity operator

    """

    return LocalOp(idx = idx,
                   map = jnp.array([j for j in range(lDim)], dtype=np.int32),
                   matEls = jnp.array([1. for j in range(lDim)], dtype=opDtype),
                   diag = True)


def Sx(idx):
    """Returns a :math:`\hat\sigma^x` Pauli operator

    Args:

    * ``idx``: Index of the local Hilbert space.

    Returns:
        Dictionary defining :math:`\hat\sigma^x` Pauli operator

    """

    return LocalOp(idx = idx,
                   map = jnp.array([1, 0], dtype=np.int32),
                   matEls = jnp.array([1.0, 1.0], dtype=opDtype),
                   diag = False)


def Sy(idx):
    """Returns a :math:`\hat\sigma^x` Pauli operator

    Args:

    * ``idx``: Index of the local Hilbert space.

    Returns:
        Dictionary defining :math:`\hat\sigma^x` Pauli operator

    """

    return LocalOp(idx = idx,
                   map = jnp.array([1, 0], dtype=np.int32),
                   matEls = jnp.array([1.j, -1.j], dtype=opDtype), 
                   diag = False)


def Sz(idx):
    """Returns a :math:`\hat\sigma^z` Pauli operator

    Args:

    * ``idx``: Index of the local Hilbert space.

    Returns:
        Dictionary defining :math:`\hat\sigma^z` Pauli operator

    """

    return LocalOp(idx = idx, 
                   map = jnp.array([0, 1], dtype=np.int32), 
                   matEls = jnp.array([-1.0, 1.0], dtype=opDtype),
                   diag = True)


def Sp(idx):
    """Returns a :math:`S^+` spin-1/2 ladder operator

    Args:

    * ``idx``: Index of the local Hilbert space.

    Returns:
        Dictionary defining :math:`S^+` ladder operator

    """

    return LocalOp(idx = idx, 
                   map = jnp.array([1, 0], dtype=np.int32), 
                   matEls = jnp.array([1.0, 0.0], dtype=opDtype),
                   diag = False)


def Sm(idx):
    """Returns a :math:`S^-` spin-1/2 ladder operator

    Args:

    * ``idx``: Index of the local Hilbert space.

    Returns:
        Dictionary defining :math:`S^-` ladder operator

    """

    return LocalOp(idx = idx, 
                   map = jnp.array([0, 0], dtype=np.int32),
                   matEls = jnp.array([0.0, 1.0], dtype=opDtype),
                   diag = False)


######################
# fermionic number operator
def number(idx):
    """Returns a :math:`c^\dagger c` fermionic number operator

    Args:

    * ``idx``: Index of the local Hilbert space.

    Returns:
        Dictionary defining :math:`c^\dagger c` fermionic number operator

    """

    return LocalOp(
        idx = idx,
        map = jax.numpy.array([0,1],dtype=np.int32),
        matEls = jax.numpy.array([0.,1.],dtype=opDtype),
        diag = True,
        fermionic = False
    )

######################
# fermionic creation operator
def creation(idx): 
    """Returns a :math:`c^\dagger` fermionic creation operator

    Args:

    * ``idx``: Index of the local Hilbert space.

    Returns:
        Dictionary defining :math:`c^\dagger` fermionic creation operator

    """
    
    return LocalOp(
        idx = idx,
        map = jax.numpy.array([1,0],dtype=np.int32),
        matEls = jax.numpy.array([1.,0.],dtype=opDtype),
        diag = False,
        fermionic = True
    )

######################
# fermionic annihilation operator
def annihilation(idx): 
    """Returns a :math:`c` fermionic creation operator

    Args:

    * ``idx``: Index of the local Hilbert space.

    Returns:
        Dictionary defining :math:`c` fermionic creation operator

    """ 
    
    return LocalOp(
        idx = idx,
        map = jax.numpy.array([1,0],dtype=np.int32),
        matEls = jax.numpy.array([0.,1.],dtype=opDtype),
        diag = False,
        fermionic = True
    )


import copy

@jax.jit
def _id_prefactor(*args, val=1.0, **kwargs):
    return val

def _prod_fun(f1, f2, *args, **kwargs):
    return f1(*args) * f2(*args)


class OpStr(tuple):
    """This class provides the interface for operator strings
    """

    def __init__(self, *args):

        super(OpStr, self).__init__()


    def __new__(cls, *args):

        factors = []
        ops = []
        for o in args:
            if isinstance(o, (LocalOp, dict)):
                ops.append(o)
            else:
                if callable(o):
                    factors.append(o)
                else:
                    factors.append(functools.partial(_id_prefactor, val=o))

        while len(factors)>1:
            factors[0] = functools.partial(_prod_fun, f1=factors[0], f2=factors.pop())

        return super(OpStr, cls).__new__(cls, tuple(factors + ops))


    def __mul__(self, other):

        if not isinstance(other, (tuple, OpStr)):
            other = OpStr(other)

        if callable(other[0]):
            return OpStr(*(other[0] * self), *(other[1:]))
        
        return OpStr(*self, *other)
    
    def __rmul__(self, a):

        if isinstance(a, dict):
            return OpStr(LocalOp(**a), *self)

        newOp = [copy.deepcopy(o) for o in self]
        if not callable(a):
            a = functools.partial(_id_prefactor, val=a)

        if callable(newOp[0]):
            newOp[0] = functools.partial(_prod_fun, f1=a, f2=newOp[0])
        else:
            newOp = [a] + newOp

        return OpStr(*tuple(newOp))


def scal_opstr(a, op):
    """Add prefactor to operator string

    Arguments:
        * ``a``: Scalar prefactor or function.
        * ``op``: Operator string.

    Returns:
        Operator string rescaled by ``a``.

    """

    if not isinstance(op, (tuple, OpStr)):
        raise RuntimeError("Can add prefactors only to OpStr or tuple objects.")
    
    if isinstance(op, tuple):
        op = OpStr(*op)

    return a * op


class LocalOp(dict):
    """This class provides the interface for operators acting on a local Hilbert space

    Initializer arguments:

        * "idx": Lattice site index,
        * "map": Indices of non-zero matrix elements,
        * "matEls": Non-zero matrix elements,
        * "diag": Boolean indicating, whether the operator is diagonal,
        * "fermionic": Boolean indicating, whether this is a fermionic operator
    """

    def __init__(self, **kwargs):
        for k in kwargs.keys():
            self[k] = kwargs[k]


    def __mul__(self, other):

        if isinstance(other, dict):
            return OpStr(self, LocalOp(**other))
        
        if isinstance(other, OpStr):
            return OpStr(self, *other)
        
        return OpStr(self, other)
    
    def __rmul__(self, other):

        return other * OpStr(self)
    

class BranchFreeOperator(Operator):
    """This class provides functionality to compute operator matrix elements

    Initializer arguments:

        * ``lDim``: Dimension of local Hilbert space.
    """

    def __init__(self, lDim=2, **kwargs):
        """Initialize ``Operator``.

        Arguments:
            * ``lDim``: Dimension of local Hilbert space.
        """
        self.ops = []
        self.lDim = lDim

        super().__init__(**kwargs)

    def add(self, opDescr):
        """Add another operator to the operator

        Arguments:
            * ``opDescr``: Operator string to be added to the operator.

        """

        self.ops.append(opDescr)
        self.compiled = False

    def __iadd__(self, opDescr):
        self.add(opDescr)
        return self

    def compile(self):
        """Compiles a operator mapping function from the given operator strings.

        """

        self.idx = []
        self.map = []
        self.matEls = []
        self.diag = []
        self.prefactor = []
        ######## fermions ########
        self.fermionic = []
        ##########################
        self.maxOpStrLength = 0
        for op in self.ops:
            tmpLen = len(op)
            if callable(op[0]):
                tmpLen -= 1
            if tmpLen > self.maxOpStrLength:
                self.maxOpStrLength = tmpLen
        IdOp = Id(lDim=self.lDim)
        o = 0
        for op in self.ops:
            self.idx.append([])
            self.map.append([])
            self.matEls.append([])
            self.fermionic.append([])
            # check whether string contains prefactor
            k0=0
            if callable(op[0]):
                self.prefactor.append((o, jax.jit(op[0])))
                k0=1
            isDiagonal = True
            for k in range(self.maxOpStrLength):
                kRev = len(op) - k - 1
                if kRev >= k0:
                    if not op[kRev]['diag']:
                        isDiagonal = False
                    self.idx[o].append(op[kRev]['idx'])
                    self.map[o].append(op[kRev]['map'])
                    self.matEls[o].append(op[kRev]['matEls'])
                    ######## fermions ########
                    fermi_check = True
                    if "fermionic" in op[kRev]:
                        if op[kRev]["fermionic"]:  
                            fermi_check = False
                            self.fermionic[o].append(1.)
                    if fermi_check:
                        self.fermionic[o].append(0.)
                    ##########################
                else:
                    self.idx[o].append(IdOp['idx'])
                    self.map[o].append(IdOp['map'])
                    self.matEls[o].append(IdOp['matEls'])
                    self.fermionic[o].append(0.)

            if isDiagonal:
                self.diag.append(o)
            o = o + 1

        self.idxC = jnp.array(self.idx, dtype=np.int32)
        self.mapC = jnp.array(self.map, dtype=np.int32)
        self.matElsC = jnp.array(self.matEls, dtype=opDtype)
        ######## fermions ########
        self.fermionicC = jnp.array(self.fermionic, dtype=np.int32)
        ##########################
        self.diag = jnp.array(self.diag, dtype=np.int32)

        def arg_fun(*args, prefactor, init):
            N = len(prefactor)
            if N<50:
                res = init
                for i,f in prefactor:
                    res[i] = f(*args)
            else:
                # parallelize this, because jit compilation for each element can be slow
                comm = MPI.COMM_WORLD
                commSize = comm.Get_size()
                rank = comm.Get_rank()
                nEls = (N + commSize - 1) // commSize
                myStart = nEls * rank
                myEnd = min(myStart+nEls, N)

                firstIdx = [0] + [prefactor[nEls * r][0]-1 for r in range(1,commSize)]
                lastIdx = [prefactor[min(nEls * (r+1), N-1)][0]-1 for r in range(commSize-1)] + [len(init)]

                res = init[firstIdx[rank]:lastIdx[rank]]

                for i,f in prefactor[myStart:myEnd]:
                    res[i-firstIdx[rank]] = f(*args)

                res = np.concatenate(comm.allgather(res), axis=0)
                
            return (jnp.array(res), )

        return functools.partial(self._get_s_primes, idxC=self.idxC, mapC=self.mapC, matElsC=self.matElsC, diag=self.diag, fermiC=self.fermionicC, prefactor=self.prefactor),\
                functools.partial(arg_fun, prefactor=self.prefactor, init=np.ones(self.idxC.shape[0], dtype=self.matElsC.dtype))

    def _get_s_primes(self, s, *args, idxC, mapC, matElsC, diag, fermiC, prefactor):

        numOps = idxC.shape[0]
        #matEl = jnp.ones(numOps, dtype=matElsC.dtype)
        matEl = args[0]
        
        sp = jnp.array([s] * numOps)

        ######## fermions ########
        dim = s.ravel().shape[0]
        mask = jnp.tril(jnp.ones((dim,dim),dtype=int),-1).T
        ##########################

        def apply_fun(c, x):
            config, configMatEl = c
            idx, sMap, matEls, fermi = x

            configShape = config.shape
            config = config.ravel()
            ######## fermions ########
            configMatEl = configMatEl * matEls[config[idx]] * jnp.prod((1 - 2 * fermi) * \
                                                                       (2 * fermi * mask[idx] +\
                                                                        (1 - 2 * fermi)) * config + \
                                                                        (1 - abs(config)))
            ##########################
            config = config.at[idx].set(sMap[config[idx]])

            return (config.reshape(configShape), configMatEl), None

        #def apply_multi(config, configMatEl, opIdx, opMap, opMatEls, prefactor):
        def apply_multi(config, configMatEl, opIdx, opMap, opMatEls, opFermi):

            (config, configMatEl), _ = jax.lax.scan(apply_fun, (config, configMatEl), (opIdx, opMap, opMatEls, opFermi))

            #return config, prefactor*configMatEl
            return config, configMatEl

        # vmap over operators
        #sp, matEl = vmap(apply_multi, in_axes=(0, 0, 0, 0, 0, 0))(sp, matEl, idxC, mapC, matElsC, jnp.array([f(*args) for f in prefactor]))
        sp, matEl = vmap(apply_multi, in_axes=(0, 0, 0, 0, 0, 0))(sp, matEl, idxC, mapC, matElsC, fermiC)
        if len(diag) > 1:
            matEl = matEl.at[diag[0]].set(jnp.sum(matEl[diag], axis=0))
            matEl = matEl.at[diag[1:]].set(jnp.zeros((diag.shape[0] - 1,), dtype=matElsC.dtype))

        return sp, matEl




import numpy as np

# Complex floating point
tCpx = np.complex128
# Real floating point
tReal = np.float64

from mpi4py import MPI

import jax

from functools import partial
import collections

try:
    myDevice = jax.devices()[MPI.COMM_WORLD.Get_rank() % len(jax.devices())]
except:
    myDevice = jax.devices()[0]
    print("WARNING: Could not assign devices based on MPI ranks. Assigning default device ", myDevice)

myPmapDevices = jax.devices()  # [myDevice]
myDeviceCount = len(myPmapDevices)
pmap_for_my_devices = partial(jax.pmap, devices=myPmapDevices)

def pmap_devices_updated(pmapDevices):
    if collections.Counter(pmapDevices) == collections.Counter(myPmapDevices):
        return False
    return True


def get_iterable(x):
    if isinstance(x, collections.abc.Iterable):
        return x
    else:
        return (x,)


def set_pmap_devices(devices):
    devices = list(get_iterable(devices))
    global myPmapDevices
    global myDeviceCount
    global pmap_for_my_devices
    myPmapDevices = devices
    myDeviceCount = len(myPmapDevices)
    pmap_for_my_devices = partial(jax.pmap, devices=myPmapDevices)
    myDevice = myPmapDevices[0]


def device_count():
    return len(myPmapDevices)


def devices():
    return myPmapDevices



import setuptools

version={}
with open("./jVMC/version.py") as fp:
    exec(fp.read(), version)

with open("README.md", "r") as fh:
    long_description = fh.read()

DEFAULT_DEPENDENCIES = ["setuptools", "wheel", "numpy", "openfermion", "jax>=0.4.12,<=0.4.31", "jaxlib>=0.4.12,<=0.4.31", "flax>=0.7.0", "mpi4py", "h5py", "PyYAML", "matplotlib", "scipy<1.13"] # Scipy version restricted, because jax is currently incompatible with new function namespace scipy.sparse.tril
#CUDA_DEPENDENCIES = ["setuptools", "wheel", "numpy", "jax[cuda]>=0.2.11,<=0.2.25", "flax>=0.3.6,<=0.3.6", "mpi4py", "h5py"]
DEV_DEPENDENCIES = DEFAULT_DEPENDENCIES + ["sphinx", "mock", "sphinx_rtd_theme", "pytest", "pytest-mpi"]

setuptools.setup(
    name='jVMC',
    version=version['__version__'],
    author="Markus Schmitt, Moritz Reh",
    author_email="markus.schmitt@uni-koeln.de",
    description="jVMC: Versatile and performant variational Monte Carlo",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://jvmc.readthedocs.io/en/latest/#",
    packages=setuptools.find_packages(),
    install_requires=DEFAULT_DEPENDENCIES,
    extras_require={
#        "cuda": CUDA_DEPENDENCIES
        "dev": DEV_DEPENDENCIES
    },
    classifiers=[
        "Programming Language :: Python :: 3",
         "License :: OSI Approved :: MIT License",
         "Operating System :: OS Independent",
    ],
)



import jax
jax.config.update("jax_enable_x64", True)
import flax
import flax.linen as nn
import numpy as np
import jax.numpy as jnp


# Find jVMC package
import sys
sys.path.append(sys.path[0] + "../..")
import jVMC
import jVMC.global_defs as global_defs
from jVMC.nets.initializers import init_fn_args
from jVMC.util.symmetries import LatticeSymmetry

from typing import Union

from functools import partial


class RNNCellStack(nn.Module):
    """
    Implementation of a stack of RNN-cells which is scanned over an input sequence.
    This is achieved by stacking multiple 'vanilla' RNN-cells to obtain a deep RNN.
    
    Arguments:
    * ``hiddenSize``: size of the hidden state vector
    * ``actFun``: non-linear activation function
    * ``initScale``: factor by which the initial parameters are scaled
    
    Returns:
    New set of hidden states (one for each layer), as well as the last hidden state, that serves as input to the output layer
    """

    cells: list
    dtype: type = global_defs.tReal
    actFun: callable = nn.elu
    initFun: callable = jax.nn.initializers.variance_scaling(scale=0.1, mode="fan_avg", distribution="uniform")

    @ nn.compact
    def __call__(self, carry, newR):
        newCarry = jnp.zeros_like(carry)
       
        newR = nn.Dense(features=carry.shape[-1], use_bias=False,
                        **init_fn_args(kernel_init=self.initFun, dtype=self.dtype), 
                        name="data_in_dense")(newR)
        newR = self.actFun(newR)

        for j, (c, cell) in enumerate(zip(carry, self.cells)):
            current_carry, newR = cell(c, newR)
            newCarry = newCarry.at[j].set(current_carry)
        return newCarry, newR

# ** end class RNNCellStack


class RNN1DGeneral(nn.Module):
    """
    Implementation of a multi-layer RNN for one-dimensional data with arbitrary cell.

    The ``cell`` parameter can be a string ("RNN", "LSTM", or "GRU") indicating a pre-implemented
    cell. Alternatively, a custom cell can be passed in the form of a tuple containing a flax
    module that implements the hidden state update rule and the initial value of the hidden state 
    (i.e., the initial ``carry``).
    The signature of the ``__call__`` function of the cell flax module has to be 
    ``(carry, state) -> (new_carry, output)``.

    This model can produce real positive or complex valued output. In either case the output is
    normalized such that

        :math:`\sum_s |RNN(s)|^{1/\kappa}=1`.

    Here, :math:`\kappa` corresponds to the initialization parameter ``logProbFactor``. Thereby, the RNN
    can represent both probability distributions and wave functions. Real or complex valued output is 
    chosen through the parameter ``realValuedOutput``.

    The RNN allows for autoregressive sampling through the ``sample`` member function.
    
    Initialization arguments:
        * ``L``: length of the spin chain
        * ``hiddenSize``: size of the hidden state vector
        * ``depth``: number of RNN-cells in the RNNCellStack
        * ``inputDim``: dimension of the input
        * ``actFun``: non-linear activation function
        * ``initScale``: factor by which the initial parameters are scaled
        * ``logProbFactor``: factor defining how output and associated sample probability are related. 0.5 for pure states and 1 for POVMs.
        * ``realValuedOutput``: Boolean indicating whether the network output is a real or complex number.
        * ``realValuedParams``: Boolean indicating whether the network parameters are real or complex parameters.
        * ``cell``: String ("RNN", "LSTM", or "GRU") or custom definition indicating which type of cell to use for hidden state  transformations.
    
    """

    L: int = 10
    hiddenSize: int = 10
    depth: int = 1
    inputDim: int = 2
    actFun: callable = nn.elu
    initScale: float = 1.0
    logProbFactor: float = 0.5
    realValuedOutput: bool = False
    realValuedParams: bool = True
    cell: Union[str, list] = "RNN"

    def setup(self):
        if isinstance(self.cell, str) and self.cell != "RNN":
            ValueError("Complex parameters for LSTM/GRU not yet implemented.")

        if self.realValuedParams:
            self.dtype = global_defs.tReal
            self.initFunction = jax.nn.initializers.variance_scaling(scale=self.initScale, mode="fan_avg", distribution="uniform", dtype=self.dtype)
        else:
            self.dtype = global_defs.tCpx
            self.initFunction = jVMC.nets.initializers.cplx_variance_scaling 

        if isinstance(self.cell, str):
            self.zero_carry = jnp.zeros((self.depth, 1, self.hiddenSize), dtype=self.dtype)
            if self.cell == "RNN":
                self.cells = [RNNCell(actFun=self.actFun, initFun=self.initFunction, dtype=self.dtype) for _ in range(self.depth)]
            elif self.cell == "LSTM":
                self.cells = [LSTMCell(features=self.hiddenSize) for _ in range(self.depth)]
                self.zero_carry = jnp.zeros((self.depth, 2, self.hiddenSize), dtype=self.dtype)
            elif self.cell == "GRU":
                self.cells = [GRUCell(features=self.hiddenSize) for _ in range(self.depth)]
            else:
                ValueError("Cell name not recognized.")
        else:
            self.cells = self.cell[0]
            self.zero_carry = self.cell[1]

        self.rnnCell = RNNCellStack(self.cells, actFun=self.actFun, dtype=self.dtype)
        init_args = init_fn_args(dtype=self.dtype, bias_init=jax.nn.initializers.zeros, kernel_init=self.initFunction)
        self.outputDense = nn.Dense(features=(self.inputDim-1) * (2 - self.realValuedOutput),
                                    use_bias=True, **init_args)

    def log_coeffs_to_log_probs(self, logCoeffs):
        phase = jnp.zeros((self.inputDim))
        if not self.realValuedOutput and self.realValuedParams:
            phase = 1.j*jnp.concatenate([jnp.array([0.0]), logCoeffs[self.inputDim-1:]]).transpose()
        amp = jnp.concatenate([jnp.array([0.0]), logCoeffs[:self.inputDim-1]]).transpose()

        return (self.logProbFactor * jax.nn.log_softmax(amp)).transpose() + phase 

    def __call__(self, x):
        _, probs = self.rnn_cell((self.zero_carry, jnp.zeros(self.inputDim)), jax.nn.one_hot(x, self.inputDim))

        return jnp.sum(probs)

    @partial(nn.transforms.scan,
             variable_broadcast='params',
             split_rngs={'params': False})
    def rnn_cell(self, carry, x):
        newCarry, out = self.rnnCell(carry[0], carry[1])
        logProb = self.log_coeffs_to_log_probs(self.outputDense(out))
        logProb = jnp.sum(logProb * x, axis=-1)
        return (newCarry, x), jnp.nan_to_num(logProb, nan=-35)

    def sample(self, batchSize, key):
        def generate_sample(key):
            myKeys = jax.random.split(key, self.L)
            _, sample = self.rnn_cell_sample(
                (self.zero_carry, jnp.zeros(self.inputDim)),
                (myKeys)
            )
            return sample[1]

        keys = jax.random.split(key, batchSize)
        return jax.vmap(generate_sample)(keys)

    @partial(nn.transforms.scan,
             variable_broadcast='params',
             split_rngs={'params': False})
    def rnn_cell_sample(self, carry, x):
        newCarry, out = self.rnnCell(carry[0], carry[1])
        logCoeffs = self.log_coeffs_to_log_probs(self.outputDense(out))
        sampleOut = jax.random.categorical(x, jnp.real(logCoeffs) / self.logProbFactor)
        return (newCarry, jax.nn.one_hot(sampleOut, self.inputDim)), (jnp.nan_to_num(logCoeffs, nan=-35), sampleOut)


class GRUCell(nn.Module):
    features: int

    @nn.compact
    def __call__(self, carry, state):
        current_carry, newR = nn.GRUCell(features=self.features, **init_fn_args(param_dtype=global_defs.tReal, recurrent_kernel_init=jax.nn.initializers.orthogonal(dtype=global_defs.tReal)))(carry, state)
        return current_carry, newR[0]


class LSTMCell(nn.Module):
    features: int

    @nn.compact
    def __call__(self, carry, state):
        current_carry, newR = nn.OptimizedLSTMCell(features=self.features, **init_fn_args(param_dtype=global_defs.tReal, recurrent_kernel_init=jax.nn.initializers.orthogonal(dtype=global_defs.tReal)))(carry, state)
        return jnp.asarray(current_carry), newR


class RNNCell(nn.Module):
    """
    Implementation of a 'vanilla' RNN-cell, that is part of an RNNCellStack which is scanned over an input sequence.
    The RNNCell therefore receives two inputs, the hidden state (if it is in a deep part of the CellStack) or the 
    input (if it is the first cell of the CellStack) aswell as the hidden state of the previous RNN-cell.
    Both inputs are mapped to obtain a new hidden state, which is what the RNNCell implements.
    
    Arguments: 
        * ``initFun``: initialization function for parameters
        * ``actFun``: non-linear activation function
        * ``dtype``: data type of parameters

    Returns:
        New hidden state
    """

    initFun: callable = jax.nn.initializers.variance_scaling(scale=1e-1, mode="fan_avg", distribution="uniform")
    actFun: callable = nn.elu
    dtype: type = global_defs.tReal

    @nn.compact
    def __call__(self, carry, state):
        cellCarry = nn.Dense(features=carry.shape[-1],
                             use_bias=False,
                             **init_fn_args(dtype=self.dtype,
                                            bias_init=jax.nn.initializers.zeros,
                                            kernel_init=self.initFun),
                             name="cell_carry_dense")

        newCarry = (self.actFun(cellCarry(carry[0])) + state)[None, :]

        return newCarry, newCarry[0]



import jax
jax.config.update("jax_enable_x64", True)
import flax
import flax.linen as nn
import jax.numpy as jnp

import jVMC.global_defs as global_defs
import jVMC.nets.activation_functions as act_funs
from jVMC.util.symmetries import LatticeSymmetry

from functools import partial
from typing import List, Sequence

import jVMC.nets.initializers
from jVMC.nets.initializers import init_fn_args


class CNN(nn.Module):
    """ Convolutional neural network with real parameters.

    Initialization arguments:
        * ``F``: Filter diameter
        * ``channels``: Number of channels
        * ``strides``: Number of pixels the filter shifts over
        * ``actFun``: Non-linear activation function
        * ``bias``: Whether to use biases
        * ``firstLayerBias``: Whether to use biases in the first layer
        * ``periodicBoundary``: Whether to use periodic boundary conditions

    """

    F: Sequence[int] = (8,)
    channels: Sequence[int] = (10,)
    strides: Sequence[int] = (1,)
    actFun: Sequence[callable] = (nn.elu,)
    bias: bool = True
    firstLayerBias: bool = False
    periodicBoundary: bool = True

    @nn.compact
    def __call__(self, x):
        initFunction = jax.nn.initializers.variance_scaling(scale=1.0, mode="fan_avg", distribution="uniform")

        # Set up padding for periodic boundary conditions
        # Padding size must be 1 - filter diameter
        pads = [(0, 0)]
        for f in self.F:
            if self.periodicBoundary:
                pads.append((0, f - 1))
            else:
                pads.append((f - 1, f - 1))
        pads.append((0, 0))

        bias = [self.bias] * len(self.channels)
        bias[0] = self.firstLayerBias

        activationFunctions = [f for f in self.actFun]
        for l in range(len(activationFunctions), len(self.channels)):
            activationFunctions.append(self.actFun[-1])

        # List of axes that will be summed for symmetrization
        reduceDims = tuple([-i - 1 for i in range(len(self.strides) + 2)])

        # init_args = init_fn_args(dtype=global_defs.tReal, kernel_init=initFunction)
        init_args = dict(dtype=global_defs.tReal, kernel_init=initFunction)

        # Add feature dimension
        x = jnp.expand_dims(jnp.expand_dims(2 * x - 1, axis=0), axis=-1)
        for c, fun, b in zip(self.channels, activationFunctions, bias):
            if self.periodicBoundary:
                x = jnp.pad(x, pads, 'wrap')
            else:
                x = jnp.pad(x, pads, 'constant', constant_values=0)

            x = fun(nn.Conv(features=c, kernel_size=tuple(self.F),
                            strides=self.strides, padding=[(0, 0)] * len(self.strides),
                            use_bias=b, **init_args)(x))

        nrm = jnp.sqrt(jnp.prod(jnp.array(x.shape[reduceDims[-1]:])))

        return jnp.sum(x, axis=reduceDims) / nrm

# ** end class CNN


class CpxCNN(nn.Module):
    """Convolutional neural network with complex parameters.

    Initialization arguments:
        * ``F``: Filter diameter
        * ``channels``: Number of channels
        * ``strides``: Number of pixels the filter shifts over
        * ``actFun``: Non-linear activation function
        * ``bias``: Whether to use biases
        * ``firstLayerBias``: Whether to use biases in the first layer
        * ``periodicBoundary``: Whether to use periodic boundary conditions

    """
    F: Sequence[int] = (8,)
    channels: Sequence[int] = (10,)
    strides: Sequence[int] = (1,)
    actFun: Sequence[callable] = (act_funs.poly6,)
    bias: bool = True
    firstLayerBias: bool = False
    periodicBoundary: bool = True

    @nn.compact
    def __call__(self, x):

        #initFunction = jax.nn.initializers.variance_scaling(scale=1.0, mode="fan_avg", distribution="uniform", dtype=global_defs.tReal)
        initFunction = jVMC.nets.initializers.cplx_variance_scaling

        # Set up padding for periodic boundary conditions
        # Padding size must be 1 - filter diameter
        pads = [(0, 0)]
        for f in self.F:
            if self.periodicBoundary:
                pads.append((0, f - 1))
            else:
                pads.append((f - 1, f - 1))
        pads.append((0, 0))

        bias = [self.bias] * len(self.channels)
        bias[0] = self.firstLayerBias

        activationFunctions = [f for f in self.actFun]
        for l in range(len(activationFunctions), len(self.channels)):
            activationFunctions.append(self.actFun[-1])
        
        # init_args = init_fn_args(dtype=global_defs.tCpx, kernel_init=initFunction)
        init_args = dict(dtype=global_defs.tCpx, kernel_init=initFunction)

        # List of axes that will be summed for symmetrization
        reduceDims = tuple([-i - 1 for i in range(len(self.strides) + 2)])

        # Add feature dimension
        x = jnp.expand_dims(jnp.expand_dims(2 * x - 1, axis=0), axis=-1)
        for c, f, b in zip(self.channels, activationFunctions, bias):
            if self.periodicBoundary:
                x = jnp.pad(x, pads, 'wrap')
            # else:
            #    x = jnp.pad(x, pads, 'constant', constant_values=0)
            x = f(nn.Conv(features=c, kernel_size=tuple(self.F),
                          strides=self.strides,
                          use_bias=b, **init_args)(x))

        # strides=self.strides, padding=[(0, 0)] * len(self.strides),
        nrm = jnp.sqrt(jnp.prod(jnp.array(x.shape[reduceDims[-1]:])))

        return jnp.sum(x, axis=reduceDims) / nrm

# ** end class CpxCNN


class CpxCNNSym(nn.Module):
    """
    Complex symmetric CNN.
    It uses the CpxCNN class to compute probabilities and averages the outputs over all symmetry-invariant configurations.

    Initialization arguments:
        * ``orbit``: orbits which define the symmetry operations (instance of ``util.symmetries.LatticeSymmetry``)
        * ``F``: Filter diameter
        * ``channels``: Number of channels
        * ``strides``: Number of pixels the filter shifts over
        * ``actFun``: Non-linear activation function
        * ``bias``: Whether to use biases
        * ``firstLayerBias``: Whether to use biases in the first layer

    """
    orbit: LatticeSymmetry
    F: Sequence[int] = (8,)
    channels: Sequence[int] = (10,)
    strides: Sequence[int] = (1,)
    actFun: Sequence[callable] = (act_funs.poly6,)
    bias: bool = True
    firstLayerBias: bool = False

    def setup(self):

        self.cnn = CpxCNN(F=self.F, channels=self.channels,
                          strides=self.strides, actFun=self.actFun,
                          bias=self.bias, firstLayerBias=self.firstLayerBias)

    def __call__(self, x):

        inShape = x.shape
        x = jax.vmap(lambda o, s: jnp.dot(o, s.ravel()).reshape(inShape), in_axes=(0, None))(self.orbit.orbit, x)

        def evaluate(x):
            return self.cnn(x)

        res = jnp.mean(jax.vmap(evaluate)(x), axis=0)

        return res

# ** end class CpxCNNSym



import unittest

import jax
jax.config.update("jax_enable_x64", True)
import jax.numpy as jnp

import numpy as np

import jVMC
import jVMC.util.symmetries as symmetries
import jVMC.nets as nets
from jVMC.vqs import NQS


class TestSymmetries(unittest.TestCase):

    def test_symmetries2D(self):
        L = 3
        for rotation in [True, False]:
            for reflection in [True, False]:
                for translation in [True, False]:
                    for spinflip in [True, False]:
                        symms = ("rotation" if rotation else None, 
                                 "reflection" if reflection else None,
                                 "translation" if translation else None,
                                 "spinflip" if spinflip else None)
                        
                        orbit = symmetries.get_orbit_2D_square(L, *symms)
                        self.assertTrue(np.issubdtype(orbit.dtype, np.integer))

    def test_symmetries1D(self):
        L = 3
        reflection_f = 2
        translation_f = L
        spinflip_f = 2
        for translation in [True, False]:
            for reflection in [True, False]:
                for spinflip in [True, False]:
                    symms = ("reflection" if reflection else None,
                             "translation" if translation else None,
                             "spinflip" if spinflip else None)
                    orbit = symmetries.get_orbit_1D(L, *symms)
                    self.assertTrue(orbit.shape[0] == (reflection_f if reflection else 1) * (translation_f if translation else 1) * (spinflip_f if spinflip else 1))
                    self.assertTrue(np.issubdtype(orbit.dtype, np.integer))


class TestSymmetrySector(unittest.TestCase):

    def test_symmetry_sector_1D(self):

        for reflection, translation, spinflip in [[True, False, False], [False, True, False], [False, False, True]]:
            L = 10
            symms = {"reflection": {"use": reflection, "factor": -1},
                     "translation": {"use": translation, "factor": jnp.exp(1j * 2 * jnp.pi * L / 2 / L)},
                     "spinflip": {"use": spinflip, "factor": -1},
                     }
            symms = ("reflection" if reflection else None,
                     "translation" if translation else None,
                     "spinflip" if spinflip else None)
            orbit = symmetries.get_orbit_1D(L, *symms,
                                              reflection_factor=-1.,
                                              translation_factor=jnp.exp(1j * 2 * jnp.pi * L / 2 / L),
                                              spinflip_factor=-1.)

            rbm = nets.CpxRBM_Nospinflip(numHidden=2, bias=False)
            net = nets.sym_wrapper.SymNet(net=rbm, orbit=orbit)

            psi = NQS(net)

            config_1 = jax.random.choice(jax.random.PRNGKey(0), jnp.array([0, 1] * (L // 2)), shape=(L,), replace=False)

            if reflection:
                config_2 = config_1[::-1]

            if translation:
                config_2 = jnp.roll(config_1, 1, axis=0)

            if spinflip:
                config_2 = 1 - config_1

            coeff_1 = psi(config_1[None, None, ...])
            coeff_2 = psi(config_2[None, None, ...])

            self.assertTrue(jnp.abs(1 + jnp.exp(coeff_1 - coeff_2)) < 1e-10)

    def test_symmetry_sector_2D(self):

        for rotation, reflection, translation, spinflip in [[True, False, False, False], [False, True, False, False], [False, False, True, False], [False, False, False, True]]:

            L = 4
            symms = ("rotation" if rotation else None, 
                     "reflection" if reflection else None,
                     "translation" if translation else None,
                     "spinflip" if spinflip else None)
            orbit = symmetries.get_orbit_2D_square(L, *symms,
                                                      rotation_factor=-1.,
                                                      reflection_factor=-1.,
                                                      translation_factor=jnp.exp(1j * 2 * jnp.pi * L / 2 / L),
                                                      spinflip_factor=-1.)

            rbm = nets.CpxRBM_Nospinflip(numHidden=2, bias=False)
            net = nets.sym_wrapper.SymNet(net=rbm, orbit=orbit)

            psi = NQS(net)

            config_1 = jax.random.choice(jax.random.PRNGKey(0), jnp.array([0, 1] * (L**2 // 2)), shape=(L, L), replace=False)

            if rotation:
                config_2 = config_1[:, ::-1].T

            if reflection:
                config_2 = config_1[::-1, :]

            if translation:
                config_2 = jnp.roll(config_1, 1, axis=0)

            if spinflip:
                config_2 = 1 - config_1

            coeff_1 = psi(config_1[None, None, ...])
            coeff_2 = psi(config_2[None, None, ...])

            self.assertTrue(jnp.abs(1 + jnp.exp(coeff_1 - coeff_2)) < 1e-10)


if __name__ == "__main__":
    unittest.main()



import h5py
import numpy as np
import jVMC.mpi_wrapper as mpi
import time

class OutputManager:
    '''This class provides functionality for I/O and timing.
    '''

    def __init__(self, dataFileName, group="/", append=False):

        self.fn = dataFileName

        self.currentGroup = "/"

        self.append = 'w'
        if append:
            self.append = 'a'

        self.set_group(group)

        self.timings = {}

    def set_group(self, group):

        if group != "/":
            self.currentGroup = "/" + group

        if mpi.rank == 0:
            with h5py.File(self.fn, self.append) as f:
                if not self.currentGroup in f:
                    f.create_group(self.currentGroup)

        self.append = 'a'

    def write_observables(self, time, **kwargs):

        if mpi.rank == 0:

            with h5py.File(self.fn, "a") as f:

                if not "observables" in f[self.currentGroup]:
                    f.create_group(self.currentGroup + "/observables")

                if not "times" in f[self.currentGroup + "/observables"]:
                    f.create_dataset(self.currentGroup + "/observables/times", (0,), maxshape=(None,), dtype='f8', chunks=True)

                newLen = len(f[self.currentGroup + "/observables/times"]) + 1
                f[self.currentGroup + "/observables/times"].resize((newLen,))
                f[self.currentGroup + "/observables/times"][-1] = time

                for key, obsDict in kwargs.items():

                    for name, value in obsDict.items():

                        datasetName = self.currentGroup + "/observables/" + key + "/" + name

                        if not key in f[self.currentGroup + "/observables"]:

                            f.create_group(self.currentGroup + "/observables/" + key)

                        if not name in f[self.currentGroup + "/observables/" + key]:

                            f.create_dataset(datasetName, (0,) + value.shape, maxshape=(None,) + value.shape, dtype='f8', chunks=True)

                        newLen = len(f[datasetName]) + 1
                        f[datasetName].resize((newLen,) + value.shape)
                        f[datasetName][-1] = value

    def write_metadata(self, time, **kwargs):

        if mpi.rank == 0:

            groupname = "metadata"

            with h5py.File(self.fn, "a") as f:

                if not groupname in f[self.currentGroup]:
                    f.create_group(self.currentGroup + "/" + groupname)

                if not "times" in f[self.currentGroup + "/" + groupname]:
                    f.create_dataset(self.currentGroup + "/" + groupname + "/times", (0,), maxshape=(None,), dtype='f8', chunks=True)

                newLen = len(f[self.currentGroup + "/" + groupname + "/times"]) + 1
                f[self.currentGroup + "/" + groupname + "/times"].resize((newLen,))
                f[self.currentGroup + "/" + groupname + "/times"][-1] = time

                for key, value in kwargs.items():

                    if not key in f[self.currentGroup + "/" + groupname]:

                        f.create_dataset(self.currentGroup + "/" + groupname + "/" + key, (0,) + value.shape, maxshape=(None,) + value.shape, dtype='f8', chunks=True)

                    newLen = len(f[self.currentGroup + "/" + groupname + "/" + key]) + 1
                    f[self.currentGroup + "/" + groupname + "/" + key].resize((newLen,) + value.shape)
                    f[self.currentGroup + "/" + groupname + "/" + key][-1] = value

    def write_network_checkpoint(self, time, weights):

        if mpi.rank == 0:

            groupname = "network_checkpoints"

            with h5py.File(self.fn, "a") as f:

                if not groupname in f[self.currentGroup]:
                    f.create_group(self.currentGroup + "/" + groupname)

                if not "times" in f[self.currentGroup + "/" + groupname]:
                    f.create_dataset(self.currentGroup + "/" + groupname + "/times", (0,), maxshape=(None,), dtype='f8', chunks=True)

                newLen = len(f[self.currentGroup + "/" + groupname + "/times"]) + 1
                f[self.currentGroup + "/" + groupname + "/times"].resize((newLen,))
                f[self.currentGroup + "/" + groupname + "/times"][-1] = time

                key = "checkpoints"
                value = weights

                if not key in f[self.currentGroup + "/" + groupname]:

                    f.create_dataset(self.currentGroup + "/" + groupname + "/" + key, (0,) + value.shape, maxshape=(None,) + value.shape, dtype='f8', chunks=True)

                newLen = len(f[self.currentGroup + "/" + groupname + "/" + key]) + 1
                f[self.currentGroup + "/" + groupname + "/" + key].resize((newLen,) + value.shape)
                f[self.currentGroup + "/" + groupname + "/" + key][-1] = value

    def get_network_checkpoint(self, time=-1, idx=-1):

        groupname = "network_checkpoints"
        key = "checkpoints"

        weights = None

        if mpi.rank == 0:

            with h5py.File(self.fn, "r") as f:
                times = np.array(f[self.currentGroup + "/" + groupname + "/times"])

            if time >= 0:
                idx = np.argmin(np.abs(times-time))

            with h5py.File(self.fn, "r") as f:
                weights = f[self.currentGroup + "/" + groupname + "/" + key][idx]
            time = times[idx]

            mpi.comm.bcast(time)

        else:

            time = mpi.comm.bcast(None)

        weights = mpi.bcast_unknown_size(weights)

        return time, weights

    def write_error_data(self, name, data, mpiRank=0):

        if mpi.rank == mpiRank:

            groupname = "error_data"

            with h5py.File(self.fn, "a") as f:

                if not groupname in f["/"]:
                    f.create_group("/" + groupname)

                f.create_dataset("/" + groupname + "/" + name, data=np.array(data))

    def write_dataset(self, name, data, groupname="/", mpiRank=0):

        if mpi.rank == mpiRank:

            with h5py.File(self.fn, "a") as f:

                if not groupname == "/":
                    if not groupname in f["/"]:
                        f.create_group("/" + groupname)

                print(data.shape)
                f.create_dataset("/" + groupname + "/" + name, data=np.array(data))

    def start_timing(self, name):

        if not name in self.timings:
            self.timings[name] = {"total": 0.0, "last_total": 0.0, "newest": 0.0, "count": 0, "init": 0.0}

        self.timings[name]["init"] = time.perf_counter()

    def stop_timing(self, name):

        toc = time.perf_counter()

        if not name in self.timings:
            self.timings[name] = {"total": 0.0, "last_total": 0.0, "newest": 0.0, "count": 0, "init": 0.0}

        elapsed = toc - self.timings[name]["init"]

        self.timings[name]["total"] += elapsed
        self.timings[name]["newest"] = elapsed
        self.timings[name]["count"] += 1

    def add_timing(self, name, elapsed):

        if not name in self.timings:
            self.timings[name] = {"total": 0.0, "last_total": 0.0, "newest": 0.0, "count": 0, "init": 0.0}

        self.timings[name]["total"] += elapsed
        self.timings[name]["newest"] = elapsed
        self.timings[name]["count"] += 1

    def print_timings(self, indent=""):

        self.print("%sRecorded timings:" % indent)
        for key, item in self.timings.items():
            self.print("%s  * %s: %fs" % (indent, key, item["total"] - item["last_total"]))

        for key in self.timings:
            self.timings[key]["last_total"] = self.timings[key]["total"]

    def print(self, text):
        if mpi.rank == 0:
            print(text, flush=True)



from mpi4py import MPI
import jax
import jax.numpy as jnp
import numpy as np

import jVMC.global_defs as global_defs

comm = MPI.COMM_WORLD
rank = comm.Get_rank()
commSize = comm.Get_size()

globNumSamples = 0
myNumSamples = 0

from functools import partial
import time
communicationTime = 0.


def _cov_helper(data, p):
    return jnp.expand_dims(
        jnp.matmul(jnp.conj(jnp.transpose(data)), jnp.multiply(p[:, None], data)),
        axis=0
    )


_sum_up_pmapd = None
_sum_sq_pmapd = None
mean_helper = None
cov_helper = None

mpiPmapDevices = None
def jit_my_stuff():
    # This is a helper function to make sure that pmap'd functions work with the actual choice of devices
    # at all times.

    global _sum_up_pmapd
    global _sum_sq_pmapd
    global mean_helper
    global cov_helper
    global mpiPmapDevices

    if global_defs.pmap_devices_updated(mpiPmapDevices):
        _sum_up_pmapd = global_defs.pmap_for_my_devices(lambda x: jax.lax.psum(jnp.sum(x, axis=0), 'i'), axis_name='i')
        # _sum_sq_pmapd = global_defs.pmap_for_my_devices(lambda data, mean, p: jax.lax.psum(jnp.conj(data - mean).dot(p[..., None] * (data - mean)), 'i'), axis_name='i', in_axes=(0, None, 0))
        _sum_sq_pmapd = global_defs.pmap_for_my_devices(lambda data, mean, p: jnp.einsum('ij, ij, i -> j', jnp.conj(data - mean[None, ...]), (data - mean[None, ...]), p), in_axes=(0, None, 0))
        mean_helper = global_defs.pmap_for_my_devices(lambda data, p: jnp.expand_dims(jnp.dot(p, data), axis=0), in_axes=(0, 0))
        cov_helper = global_defs.pmap_for_my_devices(_cov_helper, in_axes=(0, 0))

        mpiPmapDevices = global_defs.myPmapDevices


def distribute_sampling(numSamples, localDevices=None, numChainsPerDevice=1) -> int:
    """Distribute sampling tasks across processes and devices.

    For a desired total number of samples this function determines how many samples
    should be generated by each Monte Carlo chain.

    It is assumed that a given number of MC chains is running in parallel on each device,
    and that each MPI process can potentially utilize multiple devices. Since the numbers
    of samples per chain have to be identical accross the devices of one MPI process, the
    resulting total number of samples can slightly exceed the requested number of samples.

    Arguments:
        * ``numSamples``: Total number of samples.
        * ``localDevices``: Number of devices per MPI process.
        * ``numChainsPerDevice``: Number of chains run in parallel on each device.

    Returns:
        Number of samples to be generated per device to reach the desired total number of samples.
    """

    global globNumSamples

    # Determine number of samples per process
    samplesPerProcess = numSamples // commSize

    if rank < numSamples % commSize:
        samplesPerProcess += 1

    if localDevices is None:

        globNumSamples = numSamples

        return samplesPerProcess, globNumSamples

    numChainsPerProcess = localDevices * numChainsPerDevice

    def spc(spp):
        return int((spp + numChainsPerProcess - 1) // numChainsPerProcess)

    a = numSamples % commSize
    globNumSamples = (a * spc(1 + numSamples // commSize) + (commSize - a) * spc(numSamples // commSize)) * numChainsPerProcess

    return spc(samplesPerProcess), globNumSamples


def first_sample_id():

    global globNumSamples

    mySamples = globNumSamples // commSize

    firstSampleId = rank * mySamples

    if rank < globNumSamples % commSize:
        firstSampleId += rank
    else:
        firstSampleId += globNumSamples % commSize

    return firstSampleId


def global_sum(data):
    """ Computes the sum of input data across MPI processes and device/batch dimensions.

    On each MPI process the input data is assumed to be a ``jax.numpy.array`` with a leading
    device dimension followed by a batch dimension. The data is reduced by summing up along
    device and batch dimensions as well as accross MPI processes. Hence, the result is an
    array of shape ``data.shape[2:]``.

    Arguments:
        * ``data``: Array of input data.

    Returns:
        Sum of data across MPI processes and device/batch dimensions.
    """

    jit_my_stuff()

    data.block_until_ready()
    t0 = time.perf_counter()

    # Compute sum locally
    localSum = np.array(_sum_up_pmapd(data)[0])

    # Allocate memory for result
    res = np.empty_like(localSum, dtype=localSum.dtype)

    # Global sum
    comm.Allreduce(localSum, res, op=MPI.SUM)

    global communicationTime
    communicationTime += time.perf_counter() - t0

    # return jnp.array(res)
    return jax.device_put(res, global_defs.myDevice)


def global_mean(data, p):
    """ Computes the mean of input data across MPI processes and device/batch dimensions.

    On each MPI process the input data is assumed to be a ``jax.numpy.array`` with a leading
    device dimension followed by a batch dimension. The data is reduced by computing the mean
    along device and batch dimensions as well as accross MPI processes. Hence, the result is
    an array of shape ``data.shape[2:]``.

    The mean is computed using the given probabilities, i.e.,

        :math:`\\langle X\\rangle=\sum_{j=1}^{N_S} p_jX_j`

    Arguments:
        * ``data``: Array of input data.
        * ``p``: Probabilities associated with the given data.

    Returns:
        Mean of data across MPI processes and device/batch dimensions.
    """



    jit_my_stuff()

    return global_sum(mean_helper(data, p))


def global_variance(data, p):
    """ Computes the variance of input data across MPI processes and device/batch dimensions.

    On each MPI process the input data is assumed to be a ``jax.numpy.array`` with a leading
    device dimension followed by a batch dimension. The data is reduced by computing the variance
    along device and batch dimensions as well as accross MPI processes. Hence, the result is
    an array of shape ``data.shape[2:]``.

    The mean is computed using the given probabilities, i.e.,

        :math:`\\text{Var}(X)=\sum_{j=1}^{N_S} p_j |X_j-\\langle X\\rangle|^2`

    Arguments:
        * ``data``: Array of input data.
        * ``p``: Probabilities associated with the given data.

    Returns:
        Variance of data across MPI processes and device/batch dimensions.
    """

    jit_my_stuff()

    data.block_until_ready()

    mean = global_mean(data, p)
    # Compute sum locally
    localSum = np.array(_sum_sq_pmapd(data, mean, p)[0])

    # Allocate memory for result
    res = np.empty_like(localSum, dtype=localSum.dtype)

    t0 = time.perf_counter()

    # Global sum
    global globNumSamples
    comm.Allreduce(localSum, res, op=MPI.SUM)

    global communicationTime
    communicationTime += time.perf_counter() - t0

    return jax.device_put(res, global_defs.myDevice)


def global_covariance(data, p):
    """ Computes the covariance matrix of input data across MPI processes and device/batch dimensions.

    On each MPI process the input data is assumed to be a ``jax.numpy.array`` with a leading
    device dimension followed by a batch dimension and one data dimension.
    The data is reduced by computing the covariance
    matrix along device and batch dimensions as well as accross MPI processes. Hence, the result is
    an array of shape ``data.shape[2]`` :math:`\\times` ``data.shape[2]``.

    The mean is computed using the given probabilities, i.e.,

        :math:`\\text{Cov}(X)=\sum_{j=1}^{N_S} p_jX_j\\cdot X_j^\\dagger - \\bigg(\sum_{j=1}^{N_S} p_jX_j\\bigg)\\cdot\\bigg(\sum_{j=1}^{N_S}p_jX_j^\\dagger\\bigg)`

    Arguments:
        * ``data``: Array of input data.
        * ``p``: Probabilities associated with the given data.

    Returns:
        Covariance matrix of data across MPI processes and device/batch dimensions.
    """

    jit_my_stuff()

    return global_sum(cov_helper(data, p))


def bcast_unknown_size(data, root=0):
    """ Broadcast a one-dimensional array.

    This function broadcasts the input data array to all MPI processes.

    Arguments:
        * ``data``: One dimensional array of datatype ``np.float64``.
        * ``root``: Rank of root process.

    Returns:
        On each MPI process the data received from the root process.
    """

    if rank == root:
        if data.dtype != np.float64:
            raise TypeError("Datatype has to be float64.")

    dim = None
    buf = None
    if rank == root:
        dim = len(data)
        comm.bcast(dim, root=root)
        buf = np.array(data)
    else:
        dim = comm.bcast(None, root=root)
        buf = np.empty(dim, dtype=np.float64)

    comm.Bcast([buf, dim, MPI.DOUBLE], root=root)

    return buf


def gather(data):
    """ Gathers and distributes data all-to-all.
    The returned data is therefore of shape ``(commSize * data.shape[0],) + data.shape[1:]``.

    Arguments:
        * ``data``: Array of input data.

    Returns:
        Concatenated data from all devices.
    """

    res = comm.allgather(data.reshape((-1,*data.shape[2:])))
    res = np.concatenate(res)
    
    return res


def get_communication_time():
    global communicationTime
    t = communicationTime
    communicationTime = 0.
    return t



import jax
import jax.numpy as jnp
from jax.tree_util import tree_flatten, tree_unflatten, tree_map

import jVMC
import jVMC.mpi_wrapper as mpi
import jVMC.global_defs as global_defs
from jVMC.global_defs import pmap_for_my_devices

import numpy as np

from functools import partial

_mean_helper = None
_data_prep = None
_covar_helper = None
_covar_var_helper = None
_covar_data_helper = None
_trafo_helper_1 = None
_trafo_helper_2 = None
_select_helper = None
_get_subset_helper = None
_subset_mean_helper = None
_subset_data_prep = None

statsPmapDevices = None

def jit_my_stuff():
    # This is a helper function to make sure that pmap'd functions work with the actual choice of devices
    # at all times.

    global _mean_helper
    global _covar_helper
    global _covar_var_helper
    global _covar_data_helper
    global _trafo_helper_1
    global _trafo_helper_2
    global _select_helper
    global _data_prep
    global _get_subset_helper
    global _subset_mean_helper
    global _subset_data_prep

    global statsPmapDevices

    if jVMC.global_defs.pmap_devices_updated(statsPmapDevices):

        statsPmapDevices = global_defs.myPmapDevices

        _mean_helper = jVMC.global_defs.pmap_for_my_devices(lambda data, w: jnp.tensordot(w, data, axes=(0,0)), in_axes=(0, 0))
        _data_prep = jVMC.global_defs.pmap_for_my_devices(lambda data, w, mean: jax.vmap(lambda d, w, m: jnp.sqrt(w) * (d - m), in_axes=(0,0,None))(data, w, mean), in_axes=(0, 0, None))
        _covar_helper = jVMC.global_defs.pmap_for_my_devices(
                                lambda data1, data2:
                                    jnp.tensordot(
                                        jnp.conj(data1),
                                        data2, axes=(0,0)), 
                                in_axes=(0, 0)
                                )
        _covar_var_helper = jVMC.global_defs.pmap_for_my_devices(
                                    lambda data1, data2, w: 
                                        jnp.sum(
                                            jnp.abs( 
                                                jax.vmap(lambda a,b: jnp.outer(a,b))(jnp.conj(data1), data2),
                                            )**2 / w[...,None,None],
                                            axis=0),
                                    in_axes=(0, 0, 0)
                                    )
        _covar_data_helper = jVMC.global_defs.pmap_for_my_devices(lambda data1, data2, w: jax.vmap(lambda a,b,w: jnp.outer(a,b) / w)(jnp.conj(data1), data2, w), in_axes=(0, 0, 0))
        _trafo_helper_1 = jVMC.global_defs.pmap_for_my_devices(
                                lambda data, w, mean, f: f(
                                    jax.vmap(lambda x,y: x/jnp.sqrt(y), in_axes=(0,0))(data, w) 
                                    + mean
                                    ), 
                                in_axes=(0, 0, None), static_broadcasted_argnums=(3,))
        _trafo_helper_2 = jVMC.global_defs.pmap_for_my_devices(
                                lambda data, w, mean, v, f: 
                                    jnp.matmul(v, 
                                                f(
                                                jax.vmap(lambda x,y: x/jnp.sqrt(y), in_axes=(0,0))(data, w) 
                                                + mean
                                                )
                                    ), 
                                in_axes=(0, 0, None, None), static_broadcasted_argnums=(4,))
        _select_helper = jVMC.global_defs.pmap_for_my_devices( lambda ix,g: jax.vmap(lambda ix,g: g[ix], in_axes=(None, 0))(ix,g), in_axes=(None, 0) )
        _get_subset_helper = jVMC.global_defs.pmap_for_my_devices(lambda x, ixs: x[slice(*ixs)], in_axes=(0,), static_broadcasted_argnums=(1,))
        _subset_mean_helper = jVMC.global_defs.pmap_for_my_devices(lambda d, w, m: jnp.tensordot(jnp.sqrt(w), d, axes=(0,0)) + m, in_axes=(0,0,None))
        _subset_data_prep = jVMC.global_defs.pmap_for_my_devices(jax.vmap(lambda d, w, m1, m2: d+jnp.sqrt(w)*(m1-m2), in_axes=(0,0,None,None)), in_axes=(0,0,None,None))


# def get_op_estimator(psi, operator, *args):

#     op_fun = operator.compile()
#     if type(op_fun) is tuple:
#         op_fun_args = op_fun[1](*args)
#         op_fun = op_fun[0]
#     net_fun = psi.net.apply

#     def op_estimator(params, config):

#         sp, matEls = op_fun(config, *op_fun_args)

#         log_psi_s = net_fun(params, config)
#         log_psi_sp = jax.vmap(lambda s: net_fun(params,s))(sp)

#         return jnp.dot(matEls, jnp.exp(log_psi_sp - log_psi_s))

#     return op_estimator


def flat_grad(fun):

    def grad_fun(*args):
        grad_tree = jax.grad(fun)(*args)

        dtypes = [a.dtype for a in tree_flatten(args[0])[0]]
        if dtypes[0] == np.single or dtypes[0] == np.double:
            grad_vec = tree_flatten(
                        tree_map(
                            lambda x: x.ravel(), 
                            grad_tree
                            )
                        )[0]
        else:
            grad_vec = tree_flatten(
                    tree_map(
                        lambda x: [jnp.real(x.ravel()), -jnp.imag(x.ravel())], 
                        grad_tree
                        )
                    )[0]
            
        return jnp.concatenate(grad_vec)
    
    return grad_fun


class SampledObs():
    """This class implements the computation of statistics from Monte Carlo or exact samples.

    Initializer arguments:
        * ``observations``: Observations :math:`O_n` in the sample. This can be the value of an observable `O(s_n)` or the \
                plain configuration `s_n`. The array must have a leading device dimension plus a batch dimension.
        * ``weights``: Weights :math:`w_n` associated with observation :math:`O_n`.
        * ``estimator``: [optional] Function :math:`O(\\theta, s)` that computes an estimator parametrized by :math:`\\theta`
        * ``params``: [optional] A set of parameters for the estimator function.
    """

    def __init__(
            self,
            observations=None, 
            weights=None,
            estimator=None,
            params=None
            ):
        """Initializes SampledObs class.

        Args:
            * ``observations``: Observations :math:`O_n` in the sample. This can be the value of an observable `O(s_n)` or the \
                plain configuration `s_n`. The array must have a leading device dimension plus a batch dimension.
            * ``weights``: Weights :math:`w_n` associated with observation :math:`O_n`.
            * ``estimator``: [optional] Function :math:`O(\\theta, s)` that computes an estimator parametrized by :math:`\\theta`
            * ``params``: [optional] A set of parameters for the estimator function.
        """

        jit_my_stuff()

        self._weights = weights
        self._data = observations
        self._mean = None
        self._configs = None
        def estimator_not_implemented(p,s):
            raise Exception("No estimator function given.")

        self._estimator = estimator_not_implemented
        self._estimator_grad = estimator_not_implemented

        if estimator is not None:

            self._configs = observations

            self._estimator = jVMC.global_defs.pmap_for_my_devices(
                jax.vmap(lambda p, s: estimator(p,s), in_axes=(None, 0)), 
                in_axes=(None, 0)
                )

            self._estimator_grad = jVMC.global_defs.pmap_for_my_devices(
                                    jax.vmap(lambda p, s: flat_grad(lambda a, b: jnp.real(estimator(a,b)))(p,s) + 1.j*flat_grad(lambda a, b: jnp.imag(estimator(a,b)))(p,s), in_axes=(None, 0)), 
                                    in_axes=(None, 0)
                                    )
            
            if params is not None:

                observations = self._estimator(params, self._configs)

        self._compute_data_and_mean(observations)
        

    def _compute_data_and_mean(self, observations):

        if (observations is not None) and (self._weights is not None):
            if len(observations.shape) == 2:
                observations = observations[...,None]

            #self._weights = weights
            self._mean = mpi.global_sum( _mean_helper(observations,self._weights)[:, None,...]  )
            self._data = _data_prep(observations, self._weights, self._mean)


    def mean(self, params=None):
        """Returns the mean.
        """

        if params is not None:
            observations = self._estimator(params, self._configs)
            self._compute_data_and_mean(observations)

        return self._mean
    

    def mean_and_grad(self, psi, params):
        """Returns the mean and gradient of the given estimator.
        """

        obs = self._estimator(params, self._configs)
        self._compute_data_and_mean(obs)

        obsGrad = self._estimator_grad(params, self._configs)
        obsGradMean = mpi.global_sum( _mean_helper(obsGrad,self._weights)[None,...] )

        psiGrad = SampledObs( 2.0*jnp.real( psi.gradients(self._configs) ), self._weights )

        return self._mean, psiGrad.covar(self).ravel() + obsGradMean



    def covar(self, other=None):
        """Returns the covariance.

        Args:
            * ``other`` [optional]: Another instance of `SampledObs`.
        """

        if other is None:
            other = self
        
        return mpi.global_sum( _covar_helper(self._data, other._data)[:, None,...]  )
    

    def var(self):
        """Returns the variance.
        """

        return mpi.global_sum( jnp.abs(self._data)**2 )
    

    def covar_data(self, other=None):
        """Returns the covariance.

        Args:
            * ``other`` [optional]: Another instance of `SampledObs`.
        """

        if other is None:
            other = self

        return SampledObs( _covar_data_helper(self._data, other._data, self._weights), self._weights )
    

    def covar_var(self, other=None):
        """Returns the variance of the covariance.

        Args:
            * ``other`` [optional]: Another instance of `SampledObs`.
        """

        if other is None:
            other = self
        
        return mpi.global_sum( _covar_var_helper(self._data, other._data, self._weights)[:, None,...]  ) \
                    - jnp.abs(self.covar(other))**2


    def transform(self, nonLinearFun=lambda x: x, linearFun=None):
        """Returns a `SampledObs` for the transformed data.

        Args:
            * ``fun``: A function.
        """

        if linearFun is None:
            return SampledObs( _trafo_helper_1(self._data, self._weights, self._mean, nonLinearFun), self._weights )
        
        return SampledObs( _trafo_helper_2(self._data, self._weights, self._mean, linearFun, nonLinearFun), self._weights )


    def select(self, ixs):
        """Returns a `SampledObs` for the data selection indicated by the given indices.

        Args:
            * ``ixs``: Indices of selected data.
        """

        newObs = SampledObs()
        newObs._data = _select_helper(ixs, self._data)
        newObs._mean = self._mean[ixs]
        newObs._weights = self._weights

        return newObs
    
    
    def subset(self, start=None, end=None, step=None):
        """Returns a `SampledObs` for a subset of the data.

        Args:
            * ``start``: Start sample index for subset selection
            * ``end``: End sample index for subset selection
            * ``step``: Sample index step for subset selection
        """ 

        newObs = SampledObs()
        newObs._weights = _get_subset_helper(self._weights, (start, end, step))
        normalization = mpi.global_sum(newObs._weights)
        newObs._data = _get_subset_helper(self._data, (start, end, step))
        newObs._weights = newObs._weights / normalization
        newObs._data = newObs._data / jnp.sqrt(normalization)

        newObs._mean = mpi.global_sum( _subset_mean_helper(newObs._data, newObs._weights, 0.0)[:,None,...] )  + self._mean
        newObs._data = _subset_data_prep(newObs._data, newObs._weights, self._mean, newObs._mean)

        return newObs


    def tangent_kernel(self):

        all_data = mpi.gather(self._data)
        
        return jnp.matmul(all_data, jnp.conj(jnp.transpose(all_data)))





from . import global_defs
from . import nets
from . import operator
from . import util
from . import mpi_wrapper
from . import vqs
from . import sampler
from . import stats

from .version import __version__
from .global_defs import set_pmap_devices

import jax
jax.config.update("jax_enable_x64", True)
